# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Plan execution."""

import itertools
import re
from functools import reduce
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, cast

from pydantic import validate_arguments

from renku.command.command_builder import inject
from renku.core import errors
from renku.core.interface.activity_gateway import IActivityGateway
from renku.core.interface.plan_gateway import IPlanGateway
from renku.core.plugin.provider import execute
from renku.core.storage import check_external_storage, pull_paths_from_storage
from renku.core.util import communication
from renku.core.util.datetime8601 import local_now
from renku.core.util.os import is_subpath, safe_read_yaml
from renku.core.workflow.model.concrete_execution_graph import ExecutionGraph
from renku.core.workflow.plan import is_plan_removed
from renku.core.workflow.plan_factory import delete_indirect_files_list
from renku.core.workflow.value_resolution import ValueResolver
from renku.domain_model.project_context import project_context
from renku.domain_model.provenance.activity import Activity, ActivityCollection, WorkflowFileActivityCollection
from renku.domain_model.workflow.plan import AbstractPlan
from renku.domain_model.workflow.workflow_file import WorkflowFileCompositePlan

if TYPE_CHECKING:
    from networkx import DiGraph


@inject.params(activity_gateway=IActivityGateway, plan_gateway=IPlanGateway)
def execute_workflow_graph(
    dag: "DiGraph",
    activity_gateway: IActivityGateway,
    plan_gateway: IPlanGateway,
    provider="toil",
    config=None,
    workflow_file_plan: Optional[WorkflowFileCompositePlan] = None,
):
    """Execute a Run with/without subprocesses.

    Args:
        dag(DiGraph): The workflow graph to execute.
        activity_gateway(IActivityGateway): The injected activity gateway.
        plan_gateway(IPlanGateway): The injected plan gateway.
        provider: Provider to run the workflow with (Default value = "toil").
        config: Path to config for the workflow provider (Default value = None).
        workflow_file_plan (Optional[WorkflowFileCompositePlan): If passed, a workflow file is executed, so, store
            related metadata.
    """
    inputs = {i.actual_value for p in dag.nodes for i in p.inputs}
    # NOTE: Pull inputs from Git LFS or other storage backends
    if check_external_storage():
        pull_paths_from_storage(project_context.repository, *inputs)

    # check whether the none generated inputs of workflows are available
    outputs = {o.actual_value for p in dag.nodes for o in p.outputs}
    inputs = {i for i in inputs if i not in outputs and not any(is_subpath(path=i, base=o) for o in outputs)}
    for i in inputs:
        if not Path(i).exists():
            raise errors.ParameterError(f"Invalid input value: Input '{i}' does not exist!", show_prefix=False)

    delete_indirect_files_list(project_context.path)

    if config:
        config = safe_read_yaml(config)

    started_at_time = local_now()

    execute(dag=dag, basedir=project_context.path, provider=provider, config=config)

    ended_at_time = local_now()

    activities = []

    for plan in dag.nodes:
        # NOTE: Update plans are copies of Plan objects. We need to use the original Plan objects to avoid duplicates.
        original_plan = plan_gateway.get_by_id(plan.id)

        # NOTE: Workflow files don't have an original plan
        if not original_plan:
            original_plan = plan

        activity = Activity.from_plan(
            plan=plan,
            repository=project_context.repository,
            started_at_time=started_at_time,
            ended_at_time=ended_at_time,
        )
        activity.association.plan = original_plan
        activity_gateway.add(activity)
        activities.append(activity)

    if workflow_file_plan:
        activity_collection = WorkflowFileActivityCollection.from_activities(
            activities=activities, plan=workflow_file_plan
        )
        activity_gateway.add_activity_collection(activity_collection)
    elif len(activities) > 1:
        activity_collection = ActivityCollection(activities=activities)
        activity_gateway.add_activity_collection(activity_collection)


def check_for_cycles(graph: ExecutionGraph):
    """Check for cycles in the graph and raises an error if there are any."""
    if not graph.cycles:
        return

    cycles_str = []
    for cycle in graph.cycles:
        nodes = []
        for node in cycle:
            if isinstance(node, AbstractPlan):
                nodes.append(f"[{node.name}]")
            else:
                cls = node.__class__.__name__.replace("Command", "")
                nodes.append(f"{cls}: {node.actual_value}")
        cycles_str.append(" -> ".join(nodes))

    message = "Circular workflows are not supported in Renku. Please remove these cycles:\n\t"
    message += "\n\t".join(cycles_str)
    raise errors.GraphCycleError(message=message, cycles=[])


@inject.autoparams()
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def execute_workflow(
    name_or_id: str,
    set_params: List[str],
    provider: str,
    config: Optional[str],
    values: Optional[str],
    plan_gateway: IPlanGateway,
):
    """Execute a plan with specified values.

    Args:
        name_or_id(str): Name or id of the Plan to iterate.
        set_params(List[str]): List of values specified for workflow parameters.
        provider(str): Name of the workflow provider backend to use for execution.
        config(Optional[str]): Path to config for the workflow provider.
        values(Optional[str]): Path to YAMl file containing values specified for workflow parameters.
        plan_gateway(IPlanGateway): The plan gateway.
    """
    workflow = plan_gateway.get_by_name_or_id(name_or_id)

    if is_plan_removed(workflow):
        raise errors.ParameterError(f"The specified workflow '{name_or_id}' cannot be found.")

    # apply the provided parameter settings provided by user
    override_params = dict()
    if values:
        override_params.update(safe_read_yaml(values))

    if set_params:
        from deepmerge import always_merger

        for param in set_params:
            name, value = param.split("=", maxsplit=1)
            keys = name.split(".")

            set_param = reduce(lambda x, y: {y: x}, reversed(keys), value)  # type: ignore
            override_params = always_merger.merge(override_params, set_param)

    rv = ValueResolver.get(workflow, override_params)

    workflow = rv.apply()

    if rv.missing_parameters:
        communication.warn(
            f'Could not resolve the following parameters in "{workflow.name}" workflow: '
            f'{",".join(rv.missing_parameters)}'
        )

    graph = ExecutionGraph([workflow], virtual_links=True)
    execute_workflow_graph(dag=graph.workflow_graph, provider=provider, config=config)


def _extract_iterate_parameters(values: Dict[str, Any], index_pattern: re.Pattern, tag_separator: str = "@"):
    """Recursively extracts the iteration parameters from the workflow values given by the user.

    Args:
        values(Dict[str, Any]): Plan values to iterate over.
        index_pattern(re.Pattern): Pattern for parameter indexes.
        tag_separator(str, optional): Separator for tagged values (Default value = "@").

    Returns:
        Tuple of ``(iter_params, params)`` where ``params`` are regular parameters
        and ``iter_params`` are parameters with iteration values.
    """
    iter_params: Dict[str, Any] = {"indexed": {}, "params": {}, "tagged": {}}
    params: Dict[str, Any] = {}
    for param_name, param_value in values.items():
        if isinstance(param_value, str) and index_pattern.search(param_value):
            iter_params["indexed"][param_name] = param_value
            params[param_name] = param_value
        elif isinstance(param_value, list):
            if len(param_value) == 1:
                communication.warn(
                    f"The parameter '{param_name}' has only one element '{param_value}', "
                    "changing it to be a fixed parameter!"
                )
                params[param_name] = param_value[0]
                continue

            if tag_separator in param_name:
                name, tag = param_name.split(tag_separator, maxsplit=1)
                if tag in iter_params["tagged"]:
                    iter_params["tagged"][tag][name] = param_value
                else:
                    iter_params["tagged"][tag] = {name: param_value}

                params[name] = param_value
            else:
                iter_params["params"][param_name] = param_value
                params[param_name] = param_value
        elif isinstance(param_value, dict):
            inner_iter_params, inner_params = _extract_iterate_parameters(param_value, index_pattern, tag_separator)
            iter_params["params"].update([(f"{param_name}.{ik}", iv) for ik, iv in inner_iter_params["params"].items()])
            iter_params["indexed"].update(
                [(f"{param_name}.{ik}", iv) for ik, iv in inner_iter_params["indexed"].items()]
            )
            for tag, param in inner_iter_params["tagged"].items():
                if tag in iter_params["tagged"]:
                    iter_params["tagged"][tag].update([(f"{param_name}.{ik}", iv) for ik, iv in param.items()])
                else:
                    iter_params["tagged"][tag] = {f"{param_name}.{ik}": iv for ik, iv in param.items()}
            params[param_name] = inner_params
        else:
            params[param_name] = param_value
    return iter_params, params


def _validate_iterate_parameters(
    workflow: AbstractPlan, workflow_params: Dict[str, Any], iter_params: Dict[str, Any]
) -> Optional[Dict[str, Any]]:
    """Validates the user provided iteration parameters.

    Args:
        workflow(AbstractPlan): The Plan to validate parameters against.
        workflow_params(Dict[str, Any]): The plain parameters to check.
        iter_params(Dict[str, Any]): The iterative parameters to check.

    Returns:
        Dictionary of validated iteration parameters.
    """
    import copy

    rv = ValueResolver.get(copy.deepcopy(workflow), workflow_params)
    rv.apply()

    mp_paths = [mp.split(".") for mp in rv.missing_parameters]
    for collection in [iter_params["indexed"], iter_params["params"], *iter_params["tagged"].values()]:
        remove_keys = []
        for p in collection.keys():
            parameter_path = p.split(".")
            if any(parameter_path[: len(mp)] == mp for mp in mp_paths):
                remove_keys.append(p)

        for rk in remove_keys:
            collection.pop(rk)

    # validate tagged
    empty_tags = []
    for k, tagged_params in iter_params["tagged"].items():
        if len(tagged_params) == 0:
            empty_tags.append(k)
        else:
            tagged_params_values = list(tagged_params.values())
            tag_size = len(tagged_params_values[0])
            for p in tagged_params_values[1:]:
                num_params = len(p)
                if tag_size != num_params:
                    communication.error(
                        f"'{k}' tagged parameters '{tagged_params}' has different number of possible values!"
                    )
                    return None

    for et in empty_tags:
        iter_params["tagged"].pop(et)

    if (len(iter_params["indexed"]) == 0) and (len(iter_params["params"]) == 0) and (len(iter_params["tagged"]) == 0):
        raise errors.UsageError(
            "Please check the provided mappings as none of the "
            f"parameters are present in the '{workflow.name}' workflow"
        )

    if rv.missing_parameters:
        communication.confirm(
            f'Could not resolve the following parameters in "{workflow.name}" workflow: '
            f'{", ".join(rv.missing_parameters)}. Resume the execution?',
            abort=True,
        )

    return iter_params


def _build_iterations(
    workflow: AbstractPlan, workflow_params: Dict[str, Any], iter_params: Dict[str, Any], index_pattern: re.Pattern
) -> Tuple[List[AbstractPlan], List[Dict]]:
    """Instantiate the workflows for each iteration.

    Args:
        workflow(AbstractPlan): The base workflow to use as a template.
        workflow_params(Dict[str, Any]): The plain parameters to use.
        iter_params(Dict[str, Any]): The iterative parameters to use.
        index_pattern(re.Pattern): The pattern for the index placeholder.

    Returns:
        Tuple of ``(plans, itervalues)`` with ``plans`` being a list of all
        plans for each iteration and ``itervalues`` being a list of all values
        for each iteration.
    """
    import copy

    from deepmerge import always_merger

    plans = []
    execute_plan = []

    columns = list(iter_params["params"].keys())
    tagged_values = []
    for tag in iter_params["tagged"].values():
        columns.extend(tag.keys())
        tagged_values.append(zip(*tag.values()))

    def _flatten(values):
        for i in values:
            if isinstance(i, (list, tuple)):
                yield from i
            else:
                yield i

    for i, values in enumerate(itertools.product(*iter_params["params"].values(), *tagged_values)):
        plan_params = copy.deepcopy(workflow_params)
        iteration_values = {}
        for k, v in iter_params["indexed"].items():
            value = index_pattern.sub(str(i), v)
            set_param = reduce(lambda x, y: {y: x}, reversed(k.split(".")), value)  # type: ignore
            plan_params = always_merger.merge(plan_params, set_param)
            iteration_values[k] = value

        for param_key, param_value in zip(columns, _flatten(values)):
            set_param = reduce(lambda x, y: {y: x}, reversed(param_key.split(".")), param_value)  # type: ignore
            plan_params = always_merger.merge(plan_params, set_param)
            iteration_values[param_key] = param_value

        execute_plan.append(iteration_values)
        rv = ValueResolver.get(copy.deepcopy(workflow), plan_params)
        plans.append(rv.apply())

    return plans, execute_plan


@inject.autoparams()
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def iterate_workflow(
    name_or_id: str,
    mapping_path: Optional[str],
    mappings: List[str],
    dry_run: bool,
    provider: str,
    config: Optional[str],
    plan_gateway: IPlanGateway,
):
    """Iterate a workflow repeatedly with differing values.

    Args:
        name_or_id(str): Name or id of the Plan to iterate.
        mapping_path(str): Path to file defining workflow mappings.
        mappings(List[str]): List of workflow mappings.
        dry_run(bool): Whether to preview execution or actually run it.
        provider(str): Name of the workflow provider backend to use for execution.
        config(Optional[str]): Path to config for the workflow provider.
        plan_gateway(IPlanGateway): The plan gateway.
    """
    import ast

    from deepmerge import always_merger

    from renku.core.util.tabulate import tabulate

    if mapping_path is None and len(mappings) == 0:
        raise errors.UsageError("No mapping has been given for the iteration!")

    workflow = plan_gateway.get_by_name_or_id(name_or_id)

    if is_plan_removed(workflow):
        raise errors.ParameterError(f"The specified workflow '{name_or_id}' cannot be found.")

    tag_separator = "@"
    index_pattern = re.compile(r"{iter_index}")

    iter_params: Dict[str, Any] = {"indexed": {}, "params": {}, "tagged": {}}
    workflow_params = {}
    if mapping_path:
        mapping = safe_read_yaml(mapping_path)
        iter_params, workflow_params = _extract_iterate_parameters(mapping, index_pattern, tag_separator=tag_separator)

    for m in mappings:
        param_name, param_value = m.split("=", maxsplit=1)
        if index_pattern.search(param_value):
            iter_params["indexed"][param_name] = param_value  # type: ignore
        else:
            try:
                evaluated_param_value = ast.literal_eval(param_value)
            except Exception:
                raise errors.ParameterError(
                    f"The value of '{param_name}' parameter is neither a list nor templated variable!"
                )

            if isinstance(evaluated_param_value, list) and len(evaluated_param_value) == 1:
                communication.warn(
                    f"The parameter '{param_name}' has only one element '{param_value}', "
                    "changing it to be a fixed parameter!"
                )
                workflow_params[param_name] = param_value[0]
                continue
            elif not isinstance(evaluated_param_value, list):
                workflow_params[param_name] = param_value
                continue

            if tag_separator in param_name:
                name, tag = param_name.split(tag_separator, maxsplit=1)
                if tag in iter_params["tagged"]:
                    iter_params["tagged"][tag][name] = evaluated_param_value
                else:
                    iter_params["tagged"][tag] = {name: evaluated_param_value}

                param_name = name
            else:
                iter_params["params"][param_name] = evaluated_param_value

        set_param = reduce(lambda x, y: {y: x}, reversed(param_name.split(".")), param_value)  # type: ignore
        workflow_params = always_merger.merge(workflow_params, set_param)

    validated_iter_params = _validate_iterate_parameters(workflow, workflow_params, cast(Dict[str, Any], iter_params))
    if validated_iter_params is None:
        return

    plans, execute_plan = _build_iterations(workflow, workflow_params, validated_iter_params, index_pattern)

    communication.echo(f"\n\n{tabulate(execute_plan, execute_plan[0].keys())}")
    if not dry_run:
        graph = ExecutionGraph(workflows=plans, virtual_links=True)
        execute_workflow_graph(dag=graph.workflow_graph, provider=provider, config=config)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Knowledge graph building."""

import json
from typing import Dict, List, Optional, Set, Union

from pydantic import validate_arguments

from renku.command.command_builder.command import Command, inject
from renku.command.schema.activity import ActivitySchema, WorkflowFileActivityCollectionSchema
from renku.command.schema.composite_plan import CompositePlanSchema
from renku.command.schema.dataset import DatasetSchema, DatasetTagSchema
from renku.command.schema.plan import PlanSchema
from renku.command.schema.project import ProjectSchema
from renku.command.schema.workflow_file import WorkflowFileCompositePlanSchema, WorkflowFilePlanSchema
from renku.command.view_model.graph import GraphViewModel
from renku.core import errors
from renku.core.interface.activity_gateway import IActivityGateway
from renku.core.interface.database_gateway import IDatabaseGateway
from renku.core.interface.dataset_gateway import IDatasetGateway
from renku.core.interface.plan_gateway import IPlanGateway
from renku.core.interface.project_gateway import IProjectGateway
from renku.core.util.shacl import validate_graph
from renku.core.util.urls import get_host
from renku.domain_model.dataset import Dataset, DatasetTag
from renku.domain_model.project import Project
from renku.domain_model.provenance.activity import Activity, WorkflowFileActivityCollection
from renku.domain_model.workflow.composite_plan import CompositePlan
from renku.domain_model.workflow.plan import AbstractPlan, Plan
from renku.domain_model.workflow.workflow_file import WorkflowFileCompositePlan, WorkflowFilePlan

try:
    import importlib_resources  # type: ignore[import]
except ImportError:
    import importlib.resources as importlib_resources  # type: ignore


def export_graph_command():
    """Return a command for exporting graph data."""
    return Command().command(export_graph).with_database(write=False).require_migration()


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def export_graph(
    format: str = "json-ld", revision_or_range: Optional[str] = None, strict: bool = False
) -> GraphViewModel:
    """Output graph in specific format.

    Args:
        format(str, optional): Output format (Default value = "json-ld").
        revision_or_range(str, optional): Revision or range of revisions to export for (Default value = None).
        strict(bool, optional): Whether to check generated JSON-LD against the SHACL schema (Default value = False).

    Returns:
        Renku metadata as string.
    """

    format = format.lower()

    if revision_or_range:
        graph = get_graph_for_revision(revision_or_range=revision_or_range)
    else:
        graph = get_graph_for_all_objects()

    # NOTE: rewrite ids for current environment
    host = get_host()

    for node in graph:
        update_nested_node_host(node, host)

    if strict:
        _validate_graph(json.dumps(graph, indent=2), "json-ld")

    return GraphViewModel(graph)


def update_nested_node_host(node: Dict, host: str) -> None:
    """Update all @id in a node to include host if necessary.

    Args:
        node(Dict): Node to update.
        host(str): Host to set.
    """
    for k, v in node.items():
        if k == "@id":
            if not v.startswith("http") and not v.startswith("mailto"):
                node[k] = f"https://{host}{v}"
        elif isinstance(v, dict):
            update_nested_node_host(v, host)
        elif isinstance(v, list):
            for entry in v:
                if isinstance(entry, dict):
                    update_nested_node_host(entry, host)


@inject.autoparams()
def get_graph_for_revision(
    revision_or_range: str,
    database_gateway: IDatabaseGateway,
    project_gateway: IProjectGateway,
) -> List[Dict]:
    """Get the graph for changes made in a specific revision.

    Args:
        revision_or_range(str): Revision or range of revisions to export for.
        database_gateway(IDatabaseGateway): Injected database gateway.
        project_gateway(IProjectGateway): Injected project gateway.

    Returns:
        List of JSON-LD metadata.
    """
    all_objects = database_gateway.get_modified_objects_from_revision(revision_or_range=revision_or_range)

    change_types = (Project, Dataset, DatasetTag, Activity, Plan, CompositePlan)

    changed_objects = []

    for obj in all_objects:
        if isinstance(obj, change_types):
            changed_objects.append(obj)

    project = project_gateway.get_project()

    return _convert_entities_to_graph(changed_objects, project)


@inject.autoparams()
def get_graph_for_all_objects(
    project_gateway: IProjectGateway,
    dataset_gateway: IDatasetGateway,
    activity_gateway: IActivityGateway,
    plan_gateway: IPlanGateway,
) -> List[Dict]:
    """Get JSON-LD graph for all entities.

    Args:
        project_gateway(IProjectGateway): Injected project gateway.
        dataset_gateway(IDatasetGateway): Injected dataset gateway.
        activity_gateway(IActivityGateway): Injected activity gateway.
        plan_gateway(IPlanGateway): Injected plan gateway.

    Returns:
        List of JSON-LD metadata.
    """
    project = project_gateway.get_project()
    # NOTE: Include deleted activities when exporting graph
    objects: List[Union[Project, Dataset, DatasetTag, Activity, AbstractPlan, WorkflowFileActivityCollection]]

    objects = activity_gateway.get_all_activities(include_deleted=True)

    workflow_file_executions = [
        a for a in activity_gateway.get_all_activity_collections() if isinstance(a, WorkflowFileActivityCollection)
    ]
    objects.extend(workflow_file_executions)

    processed_plans = set()

    for activity in objects:
        processed_plans |= get_activity_plan_ids(activity)

    plans = [p for p in plan_gateway.get_all_plans() if p.id not in processed_plans]

    objects.extend(plans)

    objects.append(project)

    datasets = dataset_gateway.get_provenance_tails()
    objects.extend(datasets)

    for dataset in datasets:
        objects.extend(dataset_gateway.get_all_tags(dataset))

        current_dataset = dataset
        while current_dataset.is_derivation():
            assert current_dataset.derived_from is not None
            current_dataset = dataset_gateway.get_by_id(current_dataset.derived_from.value)
            objects.append(current_dataset)

    return _convert_entities_to_graph(objects, project)


def _convert_entities_to_graph(
    entities: List[Union[Project, Dataset, DatasetTag, Activity, AbstractPlan, WorkflowFileActivityCollection]],
    project: Project,
) -> List[Dict]:
    """Convert entities to JSON-LD graph.

    Args:
        entities(List[Union[Project, Dataset, DatasetTag, Activity, Plan, CompositePlan]]): Entities to convert.
        project(Project): Current project.

    Returns:
        List of JSON-LD metadata.
    """
    graph = []
    schemas = {
        Project: ProjectSchema,
        Dataset: DatasetSchema,
        DatasetTag: DatasetTagSchema,
        Activity: ActivitySchema,
        WorkflowFilePlan: WorkflowFilePlanSchema,
        Plan: PlanSchema,
        WorkflowFileCompositePlan: WorkflowFileCompositePlanSchema,
        CompositePlan: CompositePlanSchema,
        WorkflowFileActivityCollection: WorkflowFileActivityCollectionSchema,
    }

    processed_plans = set()
    project_id = project.id

    for entity in entities:
        if entity.id in processed_plans:
            continue
        if isinstance(entity, (Dataset, Activity, AbstractPlan, WorkflowFileActivityCollection)):
            # NOTE: Since the database is read-only, it's OK to modify objects; they won't be written back
            entity.unfreeze()
            entity.project_id = project_id

            if isinstance(entity, (Activity, WorkflowFileActivityCollection)):
                entity.association.plan.unfreeze()
                entity.association.plan.project_id = project_id
        schema = next(s for t, s in schemas.items() if isinstance(entity, t))
        graph.extend(schema(flattened=True).dump(entity))

        if not isinstance(entity, (Activity, WorkflowFileActivityCollection)):
            continue

        # NOTE: mark activity plans as processed
        processed_plans |= get_activity_plan_ids(entity)

    return graph


def get_activity_plan_ids(activity: Activity) -> Set[str]:
    """Get the ids of all plans associated with an activity.

    Args:
        activity(Activity): Activity tog et Plans for.

    Returns:
        Set of ids of Plans contained in Activity.
    """
    plan_stack = [activity.association.plan]
    plan_ids = set()

    while plan_stack:
        plan = plan_stack.pop()

        plan_ids.add(plan.id)

        if isinstance(plan, CompositePlan):
            plan_stack.extend(plan.plans)

    return plan_ids


def _validate_graph(rdf_graph, format):
    ref = importlib_resources.files("renku.data") / "shacl_shape.json"
    with importlib_resources.as_file(ref) as shacl_path:
        r, _, t = validate_graph(rdf_graph, shacl_path=shacl_path, format=format)

    if not r:
        raise errors.SHACLValidationError(f"{t}\nCouldn't export: Invalid Knowledge Graph data")



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Communicator classes for printing output."""

from contextlib import ExitStack, contextmanager
from functools import wraps
from threading import RLock


class CommunicationCallback:
    """Base communication callback class."""

    lock = RLock()

    def echo(self, msg, end="\n"):
        """Write a message."""

    def info(self, msg):
        """Write an info message."""

    def warn(self, msg):
        """Write a warning message."""

    def error(self, msg):
        """Write an error message."""

    def confirm(self, msg, abort=False, warning=False, default=False):
        """Get confirmation for an action."""

    def start_progress(self, name, total, **kwargs):
        """Create a new progress tracker."""

    def update_progress(self, name, amount):
        """Update a progress tracker."""

    def finalize_progress(self, name):
        """End a progress tracker."""

    def has_prompt(self):
        """Return True if communicator provides a direct prompt to users."""

    def prompt(self, msg, type=None, default=None, **kwargs):
        """Show a message prompt."""

    @contextmanager
    def busy(self, msg):
        """Indicate a busy status.

        For instance, show a spinner in the CLI.
        """
        yield


def lock_communication(method):
    """Ensure communicator is locked."""

    @wraps(method)
    def wrapper(self, *args, **kwargs):
        """Implementation of method wrapper."""
        with CommunicationCallback.lock:
            return method(self, *args, **kwargs)

    return wrapper


def ensure_communication(method):
    """Ensure communicator is enabled."""

    @wraps(method)
    def wrapper(self, *args, **kwargs):
        """Implementation of method wrapper."""
        if self._enabled:
            return method(self, *args, **kwargs)

    return wrapper


class _CommunicationManger(CommunicationCallback):
    """Manages all communication callback objects."""

    def __init__(self):
        super().__init__()
        self._listeners = []
        self._enabled = True

    @property  # type: ignore
    @lock_communication
    def listeners(self):
        """Return subscribed listeners."""
        return self._listeners.copy()

    @lock_communication
    def subscribe(self, listener):
        """Add a new listener for communications."""
        if listener not in self._listeners:
            self._listeners.append(listener)

    @lock_communication
    def unsubscribe(self, listener):
        """Remove a communication listener."""
        if listener in self._listeners:
            self._listeners.remove(listener)

    @lock_communication
    @ensure_communication
    def echo(self, msg, end="\n"):
        """Write a message."""
        for listener in self._listeners:
            listener.echo(msg, end=end)

    @lock_communication
    @ensure_communication
    def info(self, msg):
        """Write an info message."""
        for listener in self._listeners:
            listener.info(msg)

    @lock_communication
    @ensure_communication
    def warn(self, msg):
        """Write a warning message."""
        for listener in self._listeners:
            listener.warn(msg)

    @lock_communication
    @ensure_communication
    def error(self, msg):
        """Write an error message."""
        for listener in self._listeners:
            listener.error(msg)

    @lock_communication
    def has_prompt(self):
        """Return True if any communicator provides a direct prompt to users."""
        for listener in self._listeners:
            if listener.has_prompt():
                return True

    @lock_communication
    def confirm(self, msg, abort=False, warning=False, default=False):
        """Get confirmation for an action."""
        for listener in self._listeners:
            if listener.has_prompt():
                return listener.confirm(msg, abort, warning, default=default)

    @lock_communication
    def prompt(self, msg, type=None, default=None, **kwargs):
        """Show a message prompt from the first callback that has a prompt."""
        for listener in self._listeners:
            if listener.has_prompt():
                return listener.prompt(msg, type, default, **kwargs)

    @lock_communication
    @ensure_communication
    def start_progress(self, name, total, **kwargs):
        """Create a new progress tracker."""
        for listener in self._listeners:
            listener.start_progress(name, total, **kwargs)

    @lock_communication
    @ensure_communication
    def update_progress(self, name, amount):
        """Update a progress tracker."""
        for listener in self._listeners:
            listener.update_progress(name, amount)

    @lock_communication
    @ensure_communication
    def finalize_progress(self, name):
        """End a progress tracker."""
        for listener in self._listeners:
            listener.finalize_progress(name)

    @contextmanager
    def busy(self, msg):
        """Show busy indicators."""
        with ExitStack() as stack:
            for listener in self._listeners:
                stack.enter_context(listener.busy(msg))
            yield

    @lock_communication
    def disable(self):
        """Disable all outputs; by default everything is enabled."""
        self._enabled = False

    @lock_communication
    def enable(self):
        """Enable all outputs."""
        self._enabled = True


_thread_local = None


def ensure_manager(f):
    """Decorator to add communication manager to local thread storage."""

    @wraps(f)
    def wrapper(*args, **kwargs):
        global _thread_local

        if _thread_local is None:
            from werkzeug.local import Local

            _thread_local = Local()

        if getattr(_thread_local, "communication_manager", None) is None:
            _thread_local.communication_manager = _CommunicationManger()

        return f(*args, **kwargs)

    return wrapper


@ensure_manager
def subscribe(listener):
    """Subscribe a communication listener."""
    _thread_local.communication_manager.subscribe(listener)  # type: ignore[union-attr]


@ensure_manager
def unsubscribe(listener):
    """Unsubscribe a communication listener."""
    _thread_local.communication_manager.unsubscribe(listener)  # type: ignore[union-attr]


@ensure_manager
def echo(msg, end="\n"):
    """Write a message to all listeners."""
    _thread_local.communication_manager.echo(msg, end=end)  # type: ignore[union-attr]


@ensure_manager
def info(msg):
    """Write an info message to all listeners."""
    _thread_local.communication_manager.info(msg)  # type: ignore[union-attr]


@ensure_manager
def warn(msg):
    """Write a warning message to all listeners."""
    _thread_local.communication_manager.warn(msg)  # type: ignore[union-attr]


@ensure_manager
def error(msg):
    """Write an error message to all listeners."""
    _thread_local.communication_manager.error(msg)  # type: ignore[union-attr]


@ensure_manager
def has_prompt():
    """Return True if communicator provides a direct prompt to users."""
    return True


@ensure_manager
def confirm(msg, abort=False, warning=False, default=False):
    """Get confirmation for an action from all listeners."""
    return _thread_local.communication_manager.confirm(msg, abort, warning, default)  # type: ignore[union-attr]


@ensure_manager
def prompt(msg, type=None, default=None, **kwargs):
    """Show a message prompt."""
    return _thread_local.communication_manager.prompt(msg, type, default, **kwargs)  # type: ignore[union-attr]


@ensure_manager
def start_progress(name, total, **kwargs):
    """Start a progress tracker on all listeners."""
    _thread_local.communication_manager.start_progress(name, total, **kwargs)  # type: ignore[union-attr]


@ensure_manager
def update_progress(name, amount):
    """Update a progress tracker on all listeners."""
    _thread_local.communication_manager.update_progress(name, amount)  # type: ignore[union-attr]


@ensure_manager
def finalize_progress(name):
    """End a progress tracker on all listeners."""
    _thread_local.communication_manager.finalize_progress(name)  # type: ignore[union-attr]


@ensure_manager
@contextmanager
def progress(message, total: int):
    """Create a progress context manager."""

    class Progressbar:
        def __init__(self, name):
            self.name = name

        def update(self, amount: int = 1):
            update_progress(name=self.name, amount=amount)

    try:
        start_progress(name=message, total=total)
        yield Progressbar(message)
    finally:
        finalize_progress(message)


@ensure_manager
@contextmanager
def busy(msg):
    """Indicate busy status to all listeners."""
    with _thread_local.communication_manager.busy(msg):  # type: ignore[union-attr]
        yield


@ensure_manager
def get_listeners():
    """Return a list of subscribed listeners."""
    return _thread_local.communication_manager.listeners  # type: ignore[union-attr]


@ensure_manager
def disable():
    """Disable all outputs; by default everything is enabled."""
    return _thread_local.communication_manager.disable()  # type: ignore[union-attr]


@ensure_manager
def enable():
    """Enable all outputs."""
    return _thread_local.communication_manager.enable()  # type: ignore[union-attr]


__all__ = [
    "CommunicationCallback",
    "subscribe",
    "unsubscribe",
    "info",
    "warn",
    "error",
    "has_prompt",
    "confirm",
    "prompt",
    "start_progress",
    "update_progress",
    "finalize_progress",
    "disable",
    "enable",
]



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Migrate datasets with type scoped contexts for support with pyld 2.0."""

import re

from renku.core.migration.utils import OLD_METADATA_PATH, get_datasets_path


def migrate(_):
    """Migration function."""
    migrate_datasets_for_pyld2()


def migrate_datasets_for_pyld2():
    """Migrate type scoped contexts of datasets."""
    paths = get_datasets_path().rglob(OLD_METADATA_PATH)

    for path in paths:
        with path.open("r") as dataset:
            content = dataset.read()

        content = re.sub(r'"([^"])+_prov:([^"]+)":', '"\1_prov_\2":', content)
        content = re.sub(r'"([^"])+_wfprov:([^"]+)":', '"\1_wfprov_\2":', content)
        content = re.sub(r'"([^"])+_schema:([^"]+)":', '"\1_schema_\2":', content)

        with path.open("w") as dataset:
            dataset.write(content)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""JWT utilities."""

from datetime import datetime, timezone

import jwt


def is_token_expired(token: str) -> bool:
    """Return True if the given token is expired."""
    try:
        decoded_token = jwt.decode(token, options={"verify_signature": False})
    except jwt.DecodeError:
        return True

    expiration_date = decoded_token.get("exp", 0)

    # NOTE: ``datetime.utcnow`` doesn't have timezone info, so we use ``datetime.now``
    return expiration_date < datetime.now(tz=timezone.utc).timestamp()



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service templates view."""
from flask import request

from renku.ui.service.controllers.templates_read_manifest import TemplatesReadManifestCtrl
from renku.ui.service.controllers.v1.templates import TemplatesCreateProjectCtrl_v2_2, TemplatesReadManifestCtrl_v2_2
from renku.ui.service.serializers.v1.templates import ManifestTemplatesResponseRPC_1_5
from renku.ui.service.views.api_versions import V2_0, V2_1, V2_2, VERSIONS_BEFORE_2_0
from renku.ui.service.views.decorators import accepts_json, requires_cache, requires_identity
from renku.ui.service.views.error_handlers import (
    handle_common_except,
    handle_templates_create_errors,
    handle_templates_read_errors,
)


@handle_common_except
@handle_templates_read_errors
@requires_cache
@requires_identity
def read_manifest_from_template_1_5(user_data, cache):
    """
    Read templates from the manifest file of a template repository.

    ---
    get:
      description: Read templates from the manifest file of a template repository.
      parameters:
        - in: query
          name: url
          required: true
          schema:
            type: string
        - in: query
          name: ref
          schema:
            type: string
        - in: query
          name: depth
          schema:
            type: string
      responses:
        200:
          description: Listing of templates in the repository.
          content:
            application/json:
              schema: ManifestTemplatesResponseRPC
      tags:
        - templates
    """
    ctrl = TemplatesReadManifestCtrl(cache, user_data, dict(request.args))
    ctrl.RESPONSE_SERIALIZER = ManifestTemplatesResponseRPC_1_5()  # type: ignore

    return ctrl.to_response()


@handle_common_except
@handle_templates_read_errors
@requires_cache
@requires_identity
def read_manifest_from_template_v2_2(user_data, cache):
    """
    Read templates from the manifest file of a template repository.

    ---
    get:
      description: Read templates from the manifest file of a template repository.
      parameters:
        - in: query
          name: url
          required: true
          schema:
            type: string
        - in: query
          name: ref
          schema:
            type: string
        - in: query
          name: depth
          schema:
            type: string
      responses:
        200:
          description: Listing of templates in the repository.
          content:
            application/json:
              schema: ManifestTemplatesResponseRPC
      tags:
        - templates
    """
    ctrl = TemplatesReadManifestCtrl_v2_2(cache, user_data, dict(request.args))
    return ctrl.to_response()


@handle_common_except
@handle_templates_create_errors
@accepts_json
@requires_cache
@requires_identity
def create_project_from_template_v2_2(user_data, cache):
    """
    Create a new project starting using a remote template.

    ---
    post:
      description: Create a new project using a remote template.
      requestBody:
        content:
          application/json:
            schema: ProjectTemplateRequest
      responses:
        200:
          description: Details of the created project.
          content:
            application/json:
              schema: ProjectTemplateResponseRPC
      tags:
        - templates
    """
    ctrl = TemplatesCreateProjectCtrl_v2_2(cache, user_data, dict(request.json))  # type: ignore

    return ctrl.to_response()


def add_v1_specific_template_endpoints(templates_blueprint):
    """Add v1 only endpoints to blueprint."""
    templates_blueprint.route(
        "/templates.read_manifest", methods=["GET"], provide_automatic_options=False, versions=VERSIONS_BEFORE_2_0
    )(read_manifest_from_template_1_5)
    return templates_blueprint


def add_v2_specific_template_endpoints(templates_blueprint):
    """Add old v2 endpoints to blueprint."""
    templates_blueprint.route(
        "/templates.read_manifest", methods=["GET"], provide_automatic_options=False, versions=[V2_0, V2_1, V2_2]
    )(read_manifest_from_template_v2_2)
    templates_blueprint.route(
        "/templates.create_project", methods=["POST"], provide_automatic_options=False, versions=[V2_0, V2_1, V2_2]
    )(create_project_from_template_v2_2)
    return templates_blueprint



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Custom console echo."""

import functools
import os

import click

INFO = click.style("Info: ", bold=True, fg="blue")
WARNING = click.style("Warning: ", bold=True, fg="yellow")
ERROR = click.style("Error: ", bold=True, fg="red")


def echo_via_pager(*args, **kwargs):
    """Display pager only if it does not fit in one terminal screen.

    NOTE: The feature is available only on ``less``-based pager.
    """
    restore = "LESS" not in os.environ
    try:
        os.environ.setdefault("LESS", "-iXFR")
        click.echo_via_pager(*args, **kwargs)
    finally:
        if restore:
            os.environ.pop("LESS", None)


progressbar = functools.partial(
    click.progressbar, fill_char=click.style(" ", bg="green"), show_pos=True, item_show_func=lambda x: x
)


def yellow_text(text: str) -> str:
    """Return text in yellow."""
    return click.style(text, fg="yellow")


def red_text(text: str) -> str:
    """Return text in red."""
    return click.style(text, fg="red")



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service utility functions."""
from typing import Optional, overload

from renku.ui.service.config import CACHE_PROJECTS_PATH, CACHE_UPLOADS_PATH


def make_project_path(user, project):
    """Construct full path for cached project."""
    from renku.ui.service.cache.models.project import NO_BRANCH_FOLDER

    valid_user = user and "user_id" in user
    valid_project = project and "owner" in project and "name" in project and "project_id" in project

    if valid_user and valid_project:
        return (
            CACHE_PROJECTS_PATH
            / user["user_id"]
            / project["owner"]
            / project["slug"]
            / project.get("branch", NO_BRANCH_FOLDER)
        )


def make_file_path(user, cached_file):
    """Construct full path for cache file."""
    valid_user = user and "user_id" in user
    valid_file = cached_file and "file_name" in cached_file

    if valid_user and valid_file:
        return CACHE_UPLOADS_PATH / user["user_id"] / cached_file["relative_path"]


def valid_file(user, cached_file):
    """Ensure file system and cache state matches."""
    file_path = make_file_path(user, cached_file)

    if file_path.exists():
        cached_file["is_dir"] = file_path.is_dir()
        return cached_file


def new_repo_push(repo_path, source_url, source_name="origin", source_branch="master"):
    """Push a new repo to origin."""
    from renku.core.util.git import push_changes
    from renku.infrastructure.repository import Repository

    repository = Repository(repo_path)
    repository.remotes.add(source_name, source_url)
    branch = push_changes(repository, remote=source_name)
    return branch == source_branch


@overload
def normalize_git_url(git_url: None) -> None:
    ...


@overload
def normalize_git_url(git_url: str) -> str:
    ...


def normalize_git_url(git_url: Optional[str]) -> Optional[str]:
    """Remove ``.git`` postfix from a repository's url."""
    if git_url is None:
        return None

    git_url = git_url.rstrip("/")

    while git_url.lower().endswith(".git"):
        git_url = git_url[: -len(".git")]

    return git_url



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Serializers for workflows."""
import textwrap
from typing import Callable, Dict

from renku.command.format.json import dumps
from renku.command.format.tabulate import tabulate


def tabular(workflows, *, columns=None):
    """Format workflows with a tabular output.

    Args:
        workflows: Workflows to format.
        columns: Columns to show (Default value = None).

    Returns:
        Workflows in tabular format.
    """
    if not columns:
        columns = "id,name,keywords"

    _create_workflow_short_description(workflows)

    return tabulate(collection=workflows, columns=columns, columns_mapping=WORKFLOW_COLUMNS)


def _create_workflow_short_description(workflows):
    for workflow in workflows:
        lines = textwrap.wrap(workflow.description, width=64, max_lines=5) if workflow.description else []
        workflow.short_description = "\n".join(lines)


def jsonld(workflows, **kwargs):
    """Format workflow as JSON-LD.

    Args:
        workflows: Workflows to convert.

    Returns:
        JSON-LD string of workflows.
    """
    from renku.command.schema.plan import PlanSchema

    data = [PlanSchema().dump(workflow) for workflow in workflows]
    return dumps(data, indent=2)


def json(workflows, **kwargs):
    """Format workflow as JSON.

    Args:
        workflows: Workflows to convert.

    Returns:
        String of JSON data.
    """
    from renku.domain_model.workflow.plan import PlanDetailsJson

    data = [PlanDetailsJson().dump(workflow) for workflow in workflows]
    return dumps(data, indent=2)


WORKFLOW_FORMATS: Dict[str, Callable] = {
    "tabular": tabular,
    "json-ld": jsonld,
    "json": json,
}
"""Valid formatting options."""

WORKFLOW_COLUMNS = {
    "id": ("id", "id"),
    "name": ("name", None),
    "keywords": ("keywords_csv", "keywords"),
    "description": ("short_description", "description"),
    "command": ("full_command", "command"),
}

WORKFLOW_VISUALIZE_FORMATS = {
    "console": "console",
    "dot": "dot",
}



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Renku service datasets serializers for API before 2.2.

In versions before 2.2, ``renku:slug`` was referred to as ``name`` and
``schema:name`` was ``title``. From version 2.2 onward, these are called
``slug`` and ``name`` respectively.
"""

from marshmallow import Schema, ValidationError, fields, post_load

from renku.domain_model.dataset import AnnotationJson
from renku.domain_model.dataset import DatasetCreatorsJson
from renku.domain_model.dataset import DatasetCreatorsJson as DatasetCreators
from renku.domain_model.dataset import ImageObjectJson as ImageObject
from renku.domain_model.dataset import ImageObjectRequestJson as ImageObjectRequest
from renku.ui.service.serializers.common import AsyncSchema, MigrateSchema, RemoteRepositorySchema, RenkuSyncSchema
from renku.ui.service.serializers.datasets import DatasetAddFile, DatasetFileDetails
from renku.ui.service.serializers.rpc import JsonRPCResponse


class DatasetNameSchema_2_1(Schema):
    """Schema for dataset name."""

    name = fields.String(metadata={"description": "Mandatory dataset name."}, required=True, attribute="slug")


class DatasetDetails_2_1(DatasetNameSchema_2_1):
    """Schema for dataset details."""

    version = fields.String(allow_none=True)
    created_at = fields.String(allow_none=True, attribute="date_created")

    title = fields.String(attribute="name")
    creators = fields.List(fields.Nested(DatasetCreatorsJson))
    description = fields.String()
    keywords = fields.List(fields.String())
    identifier = fields.String()
    storage = fields.String()

    annotations = fields.List(fields.Nested(AnnotationJson))

    data_directory = fields.Method("get_datadir")

    @staticmethod
    def get_datadir(obj):
        """Get data directory."""
        if isinstance(obj, dict):
            return str(obj.get("datadir_path", obj.get("datadir", "")))
        if hasattr(obj, "datadir_path"):
            return obj.datadir_path

        return str(obj.get_datadir())


class DatasetDetailsRequest_2_1(DatasetDetails_2_1):
    """Request schema with dataset image information."""

    images = fields.List(fields.Nested(ImageObjectRequest))

    custom_metadata: fields.Field = fields.Dict()


class DatasetCreateRequest_2_1(AsyncSchema, DatasetDetailsRequest_2_1, RemoteRepositorySchema, MigrateSchema):
    """Request schema for a dataset create view."""

    # NOTE: Override field in DatasetDetails
    data_directory = fields.String(  # type: ignore
        load_default=None,
        metadata={"description": "Base dataset data directory. '<project.data_directory>/<dataset.name>' by default"},
    )


class DatasetCreateResponse_2_1(DatasetNameSchema_2_1, RenkuSyncSchema):
    """Response schema for a dataset create view."""


class DatasetCreateResponseRPC_2_1(JsonRPCResponse):
    """RPC response schema for dataset create view."""

    result = fields.Nested(DatasetCreateResponse_2_1)


class DatasetRemoveRequest_2_1(AsyncSchema, DatasetNameSchema_2_1, RemoteRepositorySchema, MigrateSchema):
    """Request schema for a dataset remove."""


class DatasetRemoveResponse_2_1(DatasetNameSchema_2_1, RenkuSyncSchema):
    """Response schema for a dataset create view."""


class DatasetRemoveResponseRPC_2_1(JsonRPCResponse):
    """RPC response schema for dataset create view."""

    result = fields.Nested(DatasetRemoveResponse_2_1)


class DatasetAddRequest_2_1(AsyncSchema, DatasetNameSchema_2_1, RemoteRepositorySchema, MigrateSchema):
    """Request schema for a dataset add file view."""

    files = fields.List(fields.Nested(DatasetAddFile), required=True)

    create_dataset = fields.Boolean(load_default=False)
    force = fields.Boolean(load_default=False)

    client_extras = fields.String()

    @post_load()
    def check_files(self, data, **_):
        """Check serialized file list."""
        for _file in data["files"]:
            if "file_id" in _file and "file_path" in _file:
                raise ValidationError("invalid reference found: use either `file_id` or `file_path`")

        return data


class DatasetAddResponse_2_1(DatasetNameSchema_2_1, RenkuSyncSchema):
    """Response schema for a dataset add file view."""

    project_id = fields.String(required=True)
    files = fields.List(fields.Nested(DatasetAddFile), required=True)


class DatasetAddResponseRPC_2_1(JsonRPCResponse):
    """RPC schema for a dataset add."""

    result = fields.Nested(DatasetAddResponse_2_1)


class DatasetDetailsResponse_2_1(DatasetDetails_2_1):
    """Request schema with dataset image information."""

    images = fields.List(fields.Nested(ImageObject))


class DatasetListResponse_2_1(Schema):
    """Response schema for dataset list view."""

    datasets = fields.List(fields.Nested(DatasetDetailsResponse_2_1), required=True)


class DatasetListResponseRPC_2_1(JsonRPCResponse):
    """RPC response schema for dataset list view."""

    result = fields.Nested(DatasetListResponse_2_1)


class DatasetFilesListRequest_2_1(DatasetNameSchema_2_1, RemoteRepositorySchema):
    """Request schema for dataset files list view."""


class DatasetFilesListResponse_2_1(DatasetNameSchema_2_1):
    """Response schema for dataset files list view."""

    files = fields.List(fields.Nested(DatasetFileDetails), required=True)


class DatasetFilesListResponseRPC_2_1(JsonRPCResponse):
    """RPC schema for dataset files list view."""

    result = fields.Nested(DatasetFilesListResponse_2_1)


class DatasetImportRequest_2_1(AsyncSchema, RemoteRepositorySchema, MigrateSchema):
    """Dataset import request."""

    dataset_uri = fields.String(required=True)
    name = fields.String(metadata={"description": "Optional dataset name."}, attribute="slug")
    extract = fields.Boolean()
    tag = fields.String(metadata={"description": "Dataset version to import."})
    data_directory = fields.String(
        load_default=None,
        metadata={"description": "Base dataset data directory. '<project.data_directory>/<dataset.name>' by default"},
    )


class DatasetEditRequest_2_1(
    AsyncSchema,
    DatasetDetailsRequest_2_1,
    DatasetNameSchema_2_1,
    RemoteRepositorySchema,
    MigrateSchema,
):
    """Dataset edit metadata request."""

    title = fields.String(metadata={"description": "New name of the dataset"}, attribute="name")
    description = fields.String(metadata={"description": "New description of the dataset"})
    creators = fields.List(fields.Nested(DatasetCreators), metadata={"description": "New creators of the dataset"})
    keywords = fields.List(fields.String(), allow_none=True, metadata={"description": "New keywords for the dataset"})
    images = fields.List(
        fields.Nested(ImageObjectRequest), allow_none=True, metadata={"description": "New dataset images"}
    )
    custom_metadata = fields.List(
        fields.Dict(), metadata={"description": "New custom metadata for the dataset"}, allow_none=True
    )
    custom_metadata_source = fields.String(
        allow_none=True,
        metadata={"description": "Source for the custom metadata for the dataset"},
    )


class DatasetUnlinkRequest_2_1(AsyncSchema, DatasetNameSchema_2_1, RemoteRepositorySchema, MigrateSchema):
    """Dataset unlink file request."""

    include_filters = fields.List(fields.String())
    exclude_filters = fields.List(fields.String())

    @post_load()
    def check_filters(self, data, **_):
        """Check filters."""
        include_filter = data.get("include_filters")
        exclude_filter = data.get("exclude_filters")

        if not include_filter and not exclude_filter:
            raise ValidationError("one of the filters must be specified")

        return data



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Cleanup jobs."""
import shutil

from renku.ui.service.cache import ServiceCache
from renku.ui.service.cache.models.job import USER_JOB_STATE_ENQUEUED, USER_JOB_STATE_IN_PROGRESS
from renku.ui.service.logger import worker_log


def cache_files_cleanup():
    """Cache files a cleanup job."""
    cache = ServiceCache()
    worker_log.debug("executing cache files cleanup")

    for user, files in cache.user_files():
        jobs = [
            job for job in cache.get_jobs(user) if job.state in [USER_JOB_STATE_ENQUEUED, USER_JOB_STATE_IN_PROGRESS]
        ]

        for file in files:
            if file.is_locked(jobs):
                continue

            if file.exists() and file.ttl_expired():
                worker_log.debug(f"purging file {file.file_id}:{file.file_name}")
                file.purge()
            elif not file.exists():
                file.delete()

    for user, chunks in cache.user_chunks():
        jobs = [
            job for job in cache.get_jobs(user) if job.state in [USER_JOB_STATE_ENQUEUED, USER_JOB_STATE_IN_PROGRESS]
        ]

        chunk_folders = set()

        for chunk in chunks:
            if chunk.exists() and chunk.ttl_expired():
                worker_log.debug(f"purging chunk {chunk.chunk_file_id}:{chunk.file_name}")
                chunk.purge()
                chunk_folders.add(chunk.abs_path.parent)
            elif not chunk.exists():
                chunk.delete()
                chunk_folders.add(chunk.abs_path.parent)

        for chunk_folder in chunk_folders:
            shutil.rmtree(chunk_folder, ignore_errors=True)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions for ViewModels."""

import functools
import pydoc
import re
import shutil
from typing import TYPE_CHECKING

import click

from renku.ui.cli.utils import color

if TYPE_CHECKING:
    from renku.command.view_model.composite_plan import CompositePlanViewModel
    from renku.command.view_model.plan import PlanViewModel
    from renku.command.view_model.workflow_file import StepViewModel, WorkflowFileViewModel

style_header = functools.partial(click.style, bold=True, fg=color.YELLOW)
style_key = functools.partial(click.style, bold=True, fg=color.MAGENTA)
style_value = functools.partial(click.style, bold=True)

ANSI_REGEX = re.compile(r"(?:\x1B[@-Z\\-_]|[\x80-\x9A\x9C-\x9F]|(?:\x1B\[|\x9B)[0-?]*[ -/]*[@-~])")


def strip_ansi_codes(text: str) -> str:
    """Remove ANSI formatting codes from text."""
    return ANSI_REGEX.sub("", text)


def show_text_with_pager(text: str) -> None:
    """Print text with a pager (i.e. ``less``) if appropriate.

    Args:
        text(str): The text to print.
    """
    tty_size = shutil.get_terminal_size(fallback=(120, 40))

    if len(text.splitlines()) >= tty_size.lines:
        pydoc.pipepager(text, "less --chop-long-lines -R --tilde")
    else:
        click.echo(text)


def print_markdown(text: str, err: bool = False):
    """Print markdown text to console."""
    from rich.console import Console
    from rich.markdown import Markdown

    Console(stderr=err).print(Markdown(text))


def print_key_value(key, value, print_empty: bool = True, err: bool = False, indent: int = 0):
    """Print a key value pair."""
    if print_empty or value or (not isinstance(value, bool) and value == 0):
        click.echo(" " * indent + style_key(key) + style_value(value), err=err)


def print_key(key, err: bool = False, indent: int = 0):
    """Print a key."""
    click.echo(" " * indent + style_key(key), err=err)


def print_value(value, err: bool = False, indent: int = 0):
    """Print a value."""
    click.echo(" " * indent + style_value(value), err=err)


def print_description(description, err: bool = False, indent: int = 0):
    """Print a description."""
    if description:
        indent_str = " " * indent
        click.echo(f"{indent_str}\t\t{description}", err=err)


def print_plan(plan: "PlanViewModel", err: bool = False):
    """Print a plan to stderr.

    Args:
        plan(PlanViewModel): The plan to print.
        err(bool,optional): Print to ``stderr`` (Default value = False).
    """
    print_key_value("Id: ", plan.id, err=err)
    print_key_value("Name: ", plan.name, err=err)

    if plan.description:
        print_markdown(plan.description, err=err)

    if plan.creators:
        print_key_value("Creators: ", ", ".join(str(c) for c in plan.creators), err=err)

    if plan.keywords:
        print_key_value("Keywords: ", ", ".join(k for k in plan.keywords), err=err)

    print_key_value("Command: ", plan.full_command, err=err)
    print_key_value("Success Codes: ", plan.success_codes, err=err)

    if plan.annotations:
        print_key_value("Annotations:\n", plan.annotations, err=err)

    if plan.inputs:
        print_key("Inputs:", err=err)
        for run_input in plan.inputs:
            print_value(f"\t- {run_input.name}:", err=err)
            print_description(run_input.description, err=err)
            print_key_value("\t\tDefault Value: ", run_input.default_value, err=err)
            print_key_value("\t\tPosition: ", run_input.position, print_empty=False, err=err)
            print_key_value("\t\tPrefix: ", run_input.prefix, print_empty=False, err=err)

    if plan.outputs:
        print_key("Outputs:")
        for run_output in plan.outputs:
            print_value(f"\t- {run_output.name}:", err=err)
            print_description(run_output.description, err=err)
            print_key_value("\t\tDefault Value: ", run_output.default_value, err=err)
            print_key_value("\t\tPosition: ", run_output.position, print_empty=False, err=err)
            print_key_value("\t\tPrefix: ", run_output.prefix, print_empty=False, err=err)

    if plan.parameters:
        print_key("Parameters:", err=err)
        for run_parameter in plan.parameters:
            print_value(f"\t- {run_parameter.name}:", err=err)
            print_description(run_parameter.description, err=err)
            print_key_value("\t\tDefault Value: ", run_parameter.default_value, err=err)
            print_key_value("\t\tPosition: ", run_parameter.position, print_empty=False, err=err)
            print_key_value("\t\tPrefix: ", run_parameter.prefix, print_empty=False, err=err)


def print_composite_plan(composite_plan: "CompositePlanViewModel"):
    """Print a CompositePlan to stdout."""

    print_key_value("Id: ", composite_plan.id)
    print_key_value("Name: ", composite_plan.name)

    if composite_plan.creators:
        print_key_value("Creators: ", ", ".join(str(c) for c in composite_plan.creators))

    if composite_plan.keywords:
        print_key_value("Keywords: ", ", ".join(k for k in composite_plan.keywords))

    if composite_plan.description:
        print_markdown(composite_plan.description)

    print_key("Steps:")
    for step in composite_plan.steps:
        print_value(f"\t- {step.name}:")
        print_key_value("\t\tId: ", f"{step.id}")

    if composite_plan.mappings:
        print_key("Mappings:")
        for mapping in composite_plan.mappings:
            print_value(f"\t- {mapping.name}:")

            if mapping.description:
                print_description(mapping.description)

            print_key_value("\t\tDefault Value: ", mapping.default_value)
            print_key("\t\tMaps to: ")
            for maps_to in mapping.maps_to:
                print_value(f"\t\t\t{maps_to}")

    if composite_plan.links:
        print_key("Links: ")
        for link in composite_plan.links:
            print_key_value("\t- From: ", link.source)
            print_key("\t\t To: ")
            for sink in link.sinks:
                print_value(f"\t\t\t- {sink}")


def print_workflow_file(workflow_file: "WorkflowFileViewModel"):
    """Print a workflow file to stdout."""
    print_key_value("Name: ", workflow_file.name)
    print_key_value("Path: ", workflow_file.path)

    if workflow_file.keywords:
        print_key_value("Keywords: ", ", ".join(k for k in workflow_file.keywords))

    print_key_value("Description: ", workflow_file.description, print_empty=False)

    print_key("Steps:")
    for step in workflow_file.steps:
        print_step(step)
        click.echo()


def print_step(step: "StepViewModel"):
    """Print a step of a workflow file to stderr."""
    indent = 4
    i1 = indent
    i2 = 2 * indent
    i3 = 3 * indent
    i4 = 4 * indent

    print_value(f"{step.name}:", indent=i1)

    print_key_value("Command: ", step.full_command, indent=i2)
    print_key_value("Processed command: ", step.command, indent=i2)
    print_key_value("Success codes: ", step.success_codes, print_empty=False, indent=i2)
    if step.keywords:
        print_key_value("Keywords: ", ", ".join(k for k in step.keywords), indent=i2)
    print_key_value("Description: ", step.description, print_empty=False, indent=i2)

    if step.inputs:
        print_key("Inputs:", indent=i2)
        for input in step.inputs:
            print_value(f"{input.name}:", indent=i3)
            print_key_value("Path: ", input.path, indent=i4)
            print_key_value("Prefix: ", input.prefix, print_empty=False, indent=i4)
            print_key_value("Position: ", input.position, print_empty=False, indent=i4)
            print_key_value("Description: ", input.description, print_empty=False, indent=i4)
            print_key_value("Mapped to: ", input.mapped_to, print_empty=False, indent=i4)
            print_key_value("Implicit: ", input.implicit, print_empty=False, indent=i4)

    if step.outputs:
        print_key("Outputs:", indent=i2)
        for output in step.outputs:
            print_value(f"{output.name}:", indent=i3)
            print_key_value("Path: ", output.path, indent=i4)
            print_key_value("Prefix: ", output.prefix, print_empty=False, indent=i4)
            print_key_value("Position: ", output.position, print_empty=False, indent=i4)
            print_key_value("Description: ", output.description, print_empty=False, indent=i4)
            print_key_value("Persist: ", output.persist, print_empty=False, indent=i4)
            print_key_value("Mapped to: ", output.mapped_to, print_empty=False, indent=i4)
            print_key_value("Implicit: ", output.implicit, print_empty=False, indent=i4)

    if step.parameters:
        print_key("Parameters:", indent=i2)
        for parameter in step.parameters:
            print_value(f"{parameter.name}:", indent=i3)
            print_key_value("Value: ", parameter.value, indent=i4)
            print_key_value("Prefix: ", parameter.prefix, print_empty=False, indent=i4)
            print_key_value("Position: ", parameter.position, print_empty=False, indent=i4)
            print_key_value("Description: ", parameter.description, print_empty=False, indent=i4)
            print_key_value("Implicit: ", parameter.implicit, print_empty=False, indent=i4)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Activity management."""

import itertools
import os
from collections import defaultdict
from pathlib import Path
from typing import Dict, FrozenSet, Iterable, List, NamedTuple, Optional, Set, Tuple

import networkx
from pydantic import validate_arguments

from renku.command.command_builder import inject
from renku.core import errors
from renku.core.interface.activity_gateway import IActivityGateway
from renku.core.util import communication
from renku.core.workflow.plan import get_activities, is_plan_removed, remove_plan
from renku.domain_model.entity import Entity
from renku.domain_model.project_context import project_context
from renku.domain_model.provenance.activity import Activity, Usage


def get_activities_until_paths(
    paths: List[str], sources: List[str], activity_gateway: IActivityGateway, revision: Optional[str] = None
) -> Set[Activity]:
    """Get all current activities leading to `paths`, from `sources`."""
    all_activities: Dict[str, Set[Activity]] = defaultdict(set)

    def include_newest_activity(activity):
        existing_activities = all_activities[activity.association.plan.id]
        add_activity_if_recent(activity=activity, activities=existing_activities)

    repository = project_context.repository
    commit = None

    if revision:
        commit = repository.get_commit(revision)

    for path in paths:
        checksum = None
        if commit:
            try:
                blob = commit.tree[path]
            except KeyError:
                raise errors.GitError(f"Couldn't find file {path} at revision {revision}")
            checksum = blob.hexsha

        activities = activity_gateway.get_activities_by_generation(path, checksum=checksum)

        if len(activities) == 0:
            communication.warn(f"Path '{path}' is not generated by any workflows.")
            continue

        latest_activity = max(activities, key=lambda a: a.ended_at_time)

        upstream_chains = activity_gateway.get_upstream_activity_chains(latest_activity)

        if sources:
            # NOTE: Add the activity to check if it also matches the condition
            upstream_chains.append((latest_activity,))
            # NOTE: Only include paths that is using at least one of the sources
            upstream_chains = [c for c in upstream_chains if any(u.entity.path in sources for u in c[-1].usages)]

            # NOTE: Include activity only if any of its upstream match the condition
            if upstream_chains:
                include_newest_activity(latest_activity)
        else:
            include_newest_activity(latest_activity)

        for chain in upstream_chains:
            for activity in chain:
                include_newest_activity(activity)

    return {a for activities in all_activities.values() for a in activities}


def create_activity_graph(
    activities: List[Activity],
    remove_overridden_parents=True,
    with_inputs_outputs=False,
    with_hidden_dependencies: bool = False,
) -> networkx.Graph:
    """Create a dependency DAG from activities."""
    by_usage: Dict[str, Set[Activity]] = defaultdict(set)
    by_generation: Dict[str, Set[Activity]] = defaultdict(set)

    overridden_activities: Dict[Activity, Set[str]] = defaultdict(set)

    graph = networkx.DiGraph()

    def connect_nodes_based_on_dependencies():
        for activity in activities:
            # NOTE: Make sure that activity is in the graph in case it has no connection to others
            graph.add_node(activity)
            if with_inputs_outputs:
                create_input_output_edges(activity)

            collection = (
                itertools.chain(activity.usages, activity.hidden_usages)
                if with_hidden_dependencies
                else activity.usages
            )
            for usage in collection:
                path = usage.entity.path
                by_usage[path].add(activity)
                parent_activities = by_generation[path]
                for parent in parent_activities:
                    create_edge(parent, activity, path)

            for generation in activity.generations:
                path = generation.entity.path
                by_generation[path].add(activity)
                child_activities = by_usage[path]
                for child in child_activities:
                    create_edge(activity, child, path)

    def create_input_output_edges(activity):
        for generation in activity.generations:
            path = generation.entity.path
            if not graph.has_node(path):
                graph.add_node(path)
            if not graph.has_edge(activity, path):
                graph.add_edge(activity, path)
        collection = (
            itertools.chain(activity.usages, activity.hidden_usages) if with_hidden_dependencies else activity.usages
        )
        for usage in collection:
            path = usage.entity.path
            if not graph.has_node(path):
                graph.add_node(path)
            if not graph.has_edge(path, activity):
                graph.add_edge(path, activity)

    def create_edge(parent, child, path: str):
        if with_inputs_outputs:
            if not graph.has_edge(parent, path):
                graph.add_edge(parent, path)
            if not graph.has_edge(path, child):
                graph.add_edge(path, child)
        else:
            if graph.has_edge(parent, child):
                return
            graph.add_edge(parent, child, path=path)

    def connect_nodes_by_execution_order():
        for path, values in by_generation.items():
            if len(values) <= 1:
                continue

            # NOTE: Order multiple activities that generate a common path
            create_order_among_activities(values, path)

    def create_order_among_activities(activities: Set[Activity], path):
        for a, b in itertools.combinations(activities, 2):
            if (networkx.has_path(graph, a, b) and path in overridden_activities[a]) or (
                networkx.has_path(graph, b, a) and path in overridden_activities[b]
            ):
                continue

            # NOTE: More recent activity should be executed after the other one
            # NOTE: This won't introduce a cycle in the graph because there is no other path between the two nodes
            comparison = a.compare_to(b)
            if comparison < 0:
                if not networkx.has_path(graph, a, b):
                    graph.add_edge(a, b)
                overridden_activities[a].add(path)
            elif comparison > 0:
                if not networkx.has_path(graph, b, a):
                    graph.add_edge(b, a)
                overridden_activities[b].add(path)
            else:
                raise ValueError(f"Cannot create an order between activities that generate '{path}': {a} and {b}")

    def remove_overridden_activities():
        to_be_removed = set()
        to_be_processed = set(overridden_activities.keys())

        while len(to_be_processed) > 0:
            activity = to_be_processed.pop()
            overridden_paths = overridden_activities[activity]
            generated_path = {g.entity.path for g in activity.generations}
            if generated_path != overridden_paths:
                continue

            # NOTE: All generated paths are overridden; there is no point in executing the activity
            to_be_removed.add(activity)

            if not remove_overridden_parents:
                continue

            # NOTE: Check if its parents can be removed as well
            for parent in graph.predecessors(activity):
                if parent in to_be_removed:
                    continue
                data = graph.get_edge_data(parent, activity)
                if data and "path" in data:
                    overridden_activities[parent].add(data["path"])
                    to_be_processed.add(parent)

        for activity in to_be_removed:
            graph.remove_node(activity)

    connect_nodes_based_on_dependencies()

    cycles = list(networkx.algorithms.cycles.simple_cycles(graph))

    if cycles:
        cycles = [map(lambda x: getattr(x, "id", x), cycle) for cycle in cycles]
        raise errors.GraphCycleError(cycles)

    connect_nodes_by_execution_order()
    remove_overridden_activities()

    return graph


def sort_activities(activities: List[Activity], remove_overridden_parents=True) -> List[Activity]:
    """Return a sorted list of activities based on their dependencies and execution order."""
    graph = create_activity_graph(activities, remove_overridden_parents)

    return list(networkx.topological_sort(graph))


class ModifiedActivitiesEntities(NamedTuple):
    """A class containing sets of modified/deleted activities and entities for both normal and hidden entities."""

    modified: Set[Tuple[Activity, Entity]]
    """Set of modified activity and entity tuples."""

    deleted: Set[Tuple[Activity, Entity]]
    """Set of deleted activity and entity tuples."""

    hidden_modified: Set[Tuple[Activity, Entity]]
    """Set of modified activity and entity tuples for hidden entities."""


@inject.autoparams("activity_gateway")
def get_all_modified_and_deleted_activities_and_entities(
    repository, activity_gateway: IActivityGateway, check_hidden_dependencies: bool = False
) -> ModifiedActivitiesEntities:
    """
    Return latest activities with at least one modified or deleted input along with the modified/deleted input entity.

    An activity can be repeated if more than one of its inputs are modified.

    Args:
        repository: The current ``Repository``.
        activity_gateway(IActivityGateway): The injected Activity gateway.

    Returns:
        ModifiedActivitiesEntities: Modified and deleted activities and entities.

    """
    all_activities = activity_gateway.get_all_activities()
    relevant_activities = filter_overridden_activities(all_activities)
    return get_modified_activities(
        activities=relevant_activities, repository=repository, check_hidden_dependencies=check_hidden_dependencies
    )


@inject.autoparams()
def get_downstream_generating_activities(
    starting_activities: Set[Activity],
    paths: List[str],
    ignore_deleted: bool,
    project_path: Path,
    activity_gateway: IActivityGateway,
) -> List[Activity]:
    """Return activities downstream of passed activities that generate at least a path in ``paths``.

    Args:
        starting_activities(Set[Activity]): Activities to use as starting/upstream nodes.
        paths(List[str]): Optional generated paths to end downstream chains at.
        ignore_deleted(bool): Whether to ignore deleted generations.
        project_path(Path): Path to project's root directory.
        activity_gateway(IActivityGateway): The injected Activity gateway.

    Returns:
        Set[Activity]: All activities and their downstream activities.

    """
    all_activities: Dict[str, Set[Activity]] = defaultdict(set)

    def include_newest_activity(activity):
        existing_activities = all_activities[activity.association.plan.id]
        add_activity_if_recent(activity=activity, activities=existing_activities)

    def does_activity_generate_any_paths(activity) -> bool:
        is_same = any(g.entity.path in paths for g in activity.generations)
        is_parent = any(Path(p) in Path(g.entity.path).parents for p in paths for g in activity.generations)

        return is_same or is_parent

    def has_an_existing_generation(activity) -> bool:
        for generation in activity.generations:
            if (project_path / generation.entity.path).exists():
                return True

        return False

    for starting_activity in starting_activities:
        downstream_chains = activity_gateway.get_downstream_activity_chains(starting_activity)

        if paths:
            # NOTE: Add the activity to check if it also matches the condition
            downstream_chains.append((starting_activity,))
            downstream_chains = [c for c in downstream_chains if does_activity_generate_any_paths(c[-1])]
            # NOTE: Include activity only if any of its downstream matched the condition
            include_starting_activity = len(downstream_chains) > 0
        elif ignore_deleted:  # NOTE: Excluded deleted generations only if they are not passed in ``paths``
            # NOTE: Add the activity to check if it also matches the condition
            downstream_chains.append((starting_activity,))
            downstream_chains = [c for c in downstream_chains if has_an_existing_generation(c[-1])]
            # NOTE: Include activity only if any of its downstream matched the condition
            include_starting_activity = len(downstream_chains) > 0
        else:
            include_starting_activity = True

        if include_starting_activity:
            include_newest_activity(starting_activity)

        for chain in downstream_chains:
            for activity in chain:
                if not is_activity_valid(activity):
                    # don't process further downstream activities as the plan in question was deleted
                    break
                include_newest_activity(activity)

    return list({a for activities in all_activities.values() for a in activities})


def get_modified_activities(
    activities: FrozenSet[Activity], repository, check_hidden_dependencies: bool
) -> ModifiedActivitiesEntities:
    """Get lists of activities that have modified/deleted usage entities."""

    def get_modified_activities_helper(hashes, modified, deleted, hidden: bool):
        for activity in activities:
            collection: List[Usage] = activity.hidden_usages if hidden else activity.usages  # type: ignore
            for usage in collection:
                entity = usage.entity
                current_checksum = hashes.get(entity.path, None)
                usage_path = repository.path / usage.entity.path
                if current_checksum is None or not usage_path.exists():
                    deleted.add((activity, entity))
                elif current_checksum != entity.checksum:
                    modified.add((activity, entity))

    modified: Set[Tuple[Activity, Entity]] = set()
    deleted: Set[Tuple[Activity, Entity]] = set()
    hidden_modified: Set[Tuple[Activity, Entity]] = set()

    paths = []
    hidden_paths = []

    for activity in activities:
        for usage in activity.usages:
            paths.append(usage.entity.path)
        if check_hidden_dependencies:
            for usage in activity.hidden_usages:
                hidden_paths.append(usage.entity.path)

    hashes = repository.get_object_hashes(paths=paths)
    get_modified_activities_helper(hashes=hashes, modified=modified, deleted=deleted, hidden=False)

    if check_hidden_dependencies and hidden_paths:
        hashes = repository.get_object_hashes(paths=hidden_paths)
        get_modified_activities_helper(hashes=hashes, modified=hidden_modified, deleted=set(), hidden=True)

    return ModifiedActivitiesEntities(modified=modified, deleted=deleted, hidden_modified=hidden_modified)


def filter_overridden_activities(activities: List[Activity]) -> FrozenSet[Activity]:
    """Filter out overridden activities from a list of activities."""
    relevant_activities: Dict[FrozenSet[str], Activity] = {}

    for activity in activities[::-1]:
        outputs = frozenset(g.entity.path for g in activity.generations)

        subset_of = set()
        superset_of = set()

        for o, a in relevant_activities.items():
            if outputs.issubset(o):
                subset_of.add((o, a))
            elif outputs.issuperset(o):
                superset_of.add((o, a))

        if not subset_of and not superset_of:
            relevant_activities[outputs] = activity
            continue

        if subset_of and any(activity.ended_at_time < a.ended_at_time for _, a in subset_of):
            # activity is a subset of another, newer activity, ignore it
            continue

        older_subsets = [o for o, a in superset_of if activity.ended_at_time > a.ended_at_time]

        for older_subset in older_subsets:
            # remove other activities that this activity is a superset of
            del relevant_activities[older_subset]

        relevant_activities[outputs] = activity

    return frozenset(relevant_activities.values())


def add_activity_if_recent(activity: Activity, activities: Set[Activity]):
    """Add ``activity`` to ``activities`` if it's not in the set or is the latest executed instance.

    Remove existing activities that were executed earlier.
    """
    if activity in activities:
        return

    for existing_activity in activities:
        if activity.has_identical_inputs_and_outputs_as(existing_activity):
            if activity.ended_at_time > existing_activity.ended_at_time:  # activity is newer
                activities.remove(existing_activity)
                activities.add(activity)
            return

    # NOTE: No similar activity was found
    activities.add(activity)


def get_latest_activity(activities: Iterable[Activity]) -> Optional[Activity]:
    """Return the activity that was executed after all other activities."""
    return max(activities, key=lambda a: a.ended_at_time) if activities else None


def get_latest_activity_before(activities: Iterable[Activity], activity: Activity) -> Optional[Activity]:
    """Return the latest activity that was executed before the passed activity."""
    activities_before = [a for a in activities if a.ended_at_time <= activity.ended_at_time and a.id != activity.id]
    return get_latest_activity(activities_before)


@inject.autoparams("activity_gateway")
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def revert_activity(
    *, activity_gateway: IActivityGateway, activity_id: str, delete_plan: bool, force: bool, metadata_only: bool
) -> Activity:
    """Revert an activity.

    Args:
        activity_gateway(IActivityGateway): The injected activity gateway.
        activity_id(str): ID of the activity to be reverted.
        delete_plan(bool): Delete the plan if it's not used by any other activity.
        force(bool): Revert the activity even if it has some downstream activities.
        metadata_only(bool): Only revert the metadata and don't touch generated files.

    Returns:
        The deleted activity.
    """
    repository = project_context.repository

    def delete_associated_plan(activity):
        if not delete_plan:
            return

        plan = activity.association.plan

        used_by_other_activities = any(a for a in get_activities(plan) if a.id != activity.id)
        if used_by_other_activities:
            return

        remove_plan(name_or_id=plan.id, force=True)

    def revert_generations(activity) -> Tuple[Set[str], Set[str]]:
        """Either revert each generation to an older version (created by an earlier activity) or delete it."""
        deleted_paths = set()
        updated_paths: Dict[str, str] = {}

        if metadata_only:
            return set(), set()

        for generation in activity.generations:
            path = generation.entity.path

            generator_activities = activity_gateway.get_activities_by_generation(path=path)
            generator_activities = [a for a in generator_activities if is_activity_valid(a) and not a.deleted]
            latest_generator = get_latest_activity(generator_activities)
            if latest_generator != activity:  # NOTE: A newer activity already generated the same path
                continue

            previous_generator = get_latest_activity_before(generator_activities, activity)

            if previous_generator is None:  # NOTE: The activity is the only generator
                # NOTE: Delete the path if there are no downstreams otherwise keep it
                downstream_activities = activity_gateway.get_activities_by_usage(path)
                if not downstream_activities:
                    deleted_paths.add(path)
                elif not force:
                    raise errors.ActivityDownstreamNotEmptyError(activity)
            else:  # NOTE: There is a previous generation of that path, so, revert to it
                previous_generation = next(g for g in previous_generator.generations if g.entity.path == path)
                updated_paths[path] = previous_generation.entity.checksum

        for path, previous_checksum in updated_paths.items():
            try:
                repository.copy_content_to_file(path, checksum=previous_checksum, output_path=path)
            except errors.FileNotFound:
                communication.warn(f"Cannot revert '{path}' to a previous version, will keep the current version")

        for path in deleted_paths:
            try:
                os.unlink(project_context.path / path)
            except OSError:
                communication.warn(f"Cannot delete '{path}'")

        return deleted_paths, set(updated_paths.keys())

    activity = activity_gateway.get_by_id(activity_id)

    if activity is None:
        raise errors.ParameterError(f"Cannot find activity with ID '{activity}'")
    if activity.deleted:
        raise errors.ParameterError(f"Activity with ID '{activity}' is already deleted")

    # NOTE: The order of removal is important here so don't change it
    delete_associated_plan(activity)
    revert_generations(activity)
    activity_gateway.remove(activity, force=force)
    # NOTE: Delete the activity after processing metadata or otherwise we won't see the activity as the latest generator
    activity.delete()

    return activity


def is_activity_valid(activity: Activity) -> bool:
    """Return whether this plan has not been deleted.

    Args:
        activity(Activity): The Activity whose Plan should be checked.

    Returns:
        bool: True if the activities' Plan is still valid, False otherwise.
    """
    return not is_plan_removed(plan=activity.association.plan)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Plan management."""

import itertools
import os
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any, Dict, Generator, List, Optional, Set, Tuple, Union, cast, overload

from pydantic import validate_arguments

from renku.command.command_builder import inject
from renku.command.format.workflow import WORKFLOW_FORMATS
from renku.command.view_model.activity_graph import ActivityGraphViewModel
from renku.core import errors
from renku.core.interface.activity_gateway import IActivityGateway
from renku.core.interface.plan_gateway import IPlanGateway
from renku.core.interface.project_gateway import IProjectGateway
from renku.core.plugin.workflow_file_parser import read_workflow_file
from renku.core.util import communication
from renku.core.util.git import get_git_user
from renku.core.util.os import are_paths_related, get_relative_paths
from renku.core.workflow.model.concrete_execution_graph import ExecutionGraph
from renku.core.workflow.value_resolution import CompositePlanValueResolver, ValueResolver
from renku.domain_model.constant import NO_VALUE, NoValueType
from renku.domain_model.project_context import project_context
from renku.domain_model.provenance.activity import Activity
from renku.domain_model.provenance.agent import Person
from renku.domain_model.provenance.annotation import Annotation
from renku.domain_model.workflow.composite_plan import CompositePlan
from renku.domain_model.workflow.plan import AbstractPlan, Plan
from renku.infrastructure.immutable import DynamicProxy


@inject.autoparams("plan_gateway")
def get_plan(
    plan_gateway: IPlanGateway, name_or_id_or_path: Optional[str] = None, workflow_file: Optional[str] = None
) -> Union[AbstractPlan, str]:
    """Return the latest version of a given plan in its derivative chain."""
    plan = None

    # NOTE: At first look for a plan with the given name or ID. If not found, look for a workflow file with that name.

    if name_or_id_or_path:
        try:
            wf = plan_gateway.get_by_name_or_id(name_or_id_or_path)
        except errors.WorkflowNotFoundError:
            plan = None
        else:
            plan = None if is_plan_removed(wf) else wf
    elif not workflow_file:
        raise errors.ParameterError("Either 'name_or_id_or_path' or 'workflow_file' must be passed")

    if plan:
        return plan

    path = cast(str, name_or_id_or_path or workflow_file)

    if not os.path.exists(path):
        raise errors.WorkflowNotFoundError(path)

    return path


@overload
def get_latest_plan(plan: None = ..., plan_gateway: IPlanGateway = ...) -> None:
    ...


@overload
def get_latest_plan(plan: AbstractPlan, plan_gateway: IPlanGateway = ...) -> AbstractPlan:
    ...


@inject.autoparams("plan_gateway")
def get_latest_plan(
    plan: Optional[AbstractPlan], plan_gateway: IPlanGateway
) -> Union[Optional[AbstractPlan], AbstractPlan]:
    """Return the latest version of a given plan in its derivative chain."""
    if plan is None:
        return None

    all_plans = plan_gateway.get_all_plans()

    child_plan: Optional[AbstractPlan] = plan
    while child_plan is not None:
        plan = child_plan
        child_plan = next((p for p in all_plans if p.derived_from == plan.id), None)

    return plan


@inject.autoparams()
def get_derivative_chain(
    plan: Optional[AbstractPlan], plan_gateway: IPlanGateway
) -> Generator[AbstractPlan, None, None]:
    """Return all plans in the derivative chain of a given plan including its parents/children and the plan itself."""
    if plan is None:
        return

    plan = get_latest_plan(plan)

    while plan is not None:
        yield plan
        plan = plan_gateway.get_by_id(plan.derived_from)


@inject.autoparams()
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def search_workflows(name: str, plan_gateway: IPlanGateway) -> List[str]:
    """Get all the workflows whose Plan.name start with the given name.

    Args:
        name(str): The name to search for.
        plan_gateway(IPlanGateway): Injected Plan gateway.

    Returns:
        All Plans whose name starts with ``name``.
    """
    return plan_gateway.list_by_name(starts_with=name)


@inject.autoparams()
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def list_workflows(plan_gateway: IPlanGateway, format: str, columns: str):
    """List or manage workflows with subcommands.

    Args:
        plan_gateway(IPlanGateway): The injected Plan gateway.
        format(str): The output format.
        columns(List[str]): The columns to show for tabular output.

    Returns:
        List of workflows formatted by ``format``.
    """
    from renku.command.view_model.plan import plan_view

    workflows = plan_gateway.get_newest_plans_by_names()

    if format not in WORKFLOW_FORMATS:
        raise errors.UsageError(f'Provided format "{format}" is not supported ({", ".join(WORKFLOW_FORMATS.keys())})"')

    if format == "json-ld":
        return WORKFLOW_FORMATS[format](list(workflows.values()), columns=columns)

    return WORKFLOW_FORMATS[format](list(map(lambda x: plan_view(x), workflows.values())), columns=columns)


@inject.autoparams("activity_gateway")
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def show_workflow(name_or_id_or_path: str, activity_gateway: IActivityGateway, with_metadata: bool = False):
    """Show the details of a workflow.

    Args:
        name_or_id_or_path(str): Name or id of the Plan to show or path to a workflow file.
        activity_gateway(IActivityGateway): The injected Activity gateway.
        with_metadata(bool): Whether to get additional calculated metadata for the plan.
    Returns:
        Details of the Plan.
    """
    from renku.command.view_model.plan import plan_view
    from renku.command.view_model.workflow_file import WorkflowFileViewModel

    plan = get_plan(name_or_id_or_path=name_or_id_or_path)

    if isinstance(plan, str):  # A workflow file
        workflow_file = read_workflow_file(path=plan)
        return WorkflowFileViewModel.from_workflow_file(workflow_file)
    else:
        workflow = plan

        if with_metadata:
            activities = activity_gateway.get_all_activities()
            activity_map = _reverse_activity_plan_map(activities)
            plan_chain = list(get_derivative_chain(workflow))
            relevant_activities = [a for p in plan_chain for a in activity_map.get(p.id, [])]
            touches_files_cache: Dict[str, bool] = {}
            duration_cache: Dict[str, Optional[timedelta]] = {}
            touches_existing_files = _check_workflow_touches_existing_files(workflow, touches_files_cache, activity_map)

            if isinstance(workflow, Plan):
                num_executions = 0
                last_execution: Optional[datetime] = None

                for activity in relevant_activities:
                    num_executions += 1

                    if last_execution is None or last_execution < activity.ended_at_time:
                        last_execution = activity.ended_at_time

                workflow = cast(Plan, DynamicProxy(workflow))
                workflow.number_of_executions = num_executions
                workflow.last_executed = last_execution
                workflow.touches_existing_files = touches_existing_files
                workflow.latest = plan_chain[0].id
                workflow.duration = _get_plan_duration(workflow, duration_cache, activity_map)
            else:
                workflow = cast(CompositePlan, DynamicProxy(workflow))
                workflow.touches_existing_files = touches_existing_files
                workflow.latest = plan_chain[0].id
                workflow.duration = _get_plan_duration(workflow, duration_cache, activity_map)
                workflow.newest_plans = [get_latest_plan(p) for p in workflow.plans]

        return plan_view(cast(AbstractPlan, workflow), latest=with_metadata)


@inject.autoparams("plan_gateway")
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def remove_plan(name_or_id: str, force: bool, plan_gateway: IPlanGateway):
    """Remove the workflow by its name or id.

    Args:
        name_or_id (str): The name of the Plan to remove.
        force (bool): Whether to force removal or not.
        plan_gateway(IPlanGateway): The injected Plan gateway.
        when(datetime): Time of deletion (Default value = current local date/time).
    Raises:
        errors.ParameterError: If the Plan doesn't exist or was already deleted.
    """
    plan: Optional[AbstractPlan] = plan_gateway.get_by_name(name_or_id) or plan_gateway.get_by_id(name_or_id)

    if not plan:
        raise errors.ParameterError(f"The specified workflow '{name_or_id}' cannot be found.")

    latest_version = get_latest_plan(plan)

    if latest_version.deleted:
        raise errors.ParameterError(f"The specified workflow '{name_or_id}' is already deleted.")

    composites_containing_child = get_composite_plans_by_child(plan)

    if composites_containing_child:
        composite_names = "\n\t".join([c.name for c in composites_containing_child])

        if not force:
            raise errors.ParameterError(
                f"The specified workflow '{name_or_id}' is part of the following composite workflows and won't be "
                f"removed (use '--force' to remove anyways):\n\t{composite_names}"
            )
        else:
            communication.warn(f"Removing '{name_or_id}', which is still used in these workflows:\n\t{composite_names}")

    if not force:
        prompt_text = f"You are about to remove the following workflow '{name_or_id}'.\n\nDo you wish to continue?"
        communication.confirm(prompt_text, abort=True, warning=True)

    derived_plan = latest_version.derive()
    derived_plan.delete()

    plan_gateway.add(derived_plan)


@inject.autoparams()
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def edit_workflow(
    name: str,
    new_name: Optional[str],
    description: Optional[str],
    set_params: List[str],
    map_params: List[str],
    rename_params: List[str],
    describe_params: List[str],
    creators: Union[List[Person], NoValueType],
    keywords: Union[List[str], NoValueType],
    plan_gateway: IPlanGateway,
    custom_metadata: Optional[Dict] = None,
):
    """Edits a workflow details.

    Args:
        name (str): Name of the Plan to edit.
        new_name(Optional[str]): New name of the Plan.
        description(Optional[str]): New description of the Plan.
        set_params(List[str]): New default values for parameters.
        map_params(List[str]): New mappings for Plan.
        rename_params(List[str]): New names for parameters.
        describe_params(List[str]): New descriptions for parameters.
        creators(Union[List[Person], NoValueType]): Creators of the workflow.
        keywords(Union[List[str], NoValueType]): New keywords for the workflow.
        plan_gateway(IPlanGateway): Injected plan gateway.
        custom_metadata(Dict, optional): Custom JSON-LD metadata (Default value = None).

    Returns:
        Details of the modified Plan.
    """
    from renku.command.view_model.plan import plan_view

    derived_from = plan_gateway.get_by_name_or_id(name)

    if is_plan_removed(derived_from):
        raise errors.ParameterError(f"The specified workflow '{name}' cannot be found.")

    if (
        not new_name
        and not description
        and not set_params
        and not map_params
        and not rename_params
        and not describe_params
        and not custom_metadata
        and creators == NO_VALUE
        and keywords == NO_VALUE
    ):
        # NOTE: Nothing to do
        return plan_view(derived_from)

    git_creator = cast(Person, get_git_user(project_context.repository))
    workflow = derived_from.derive(creator=git_creator)
    if new_name:
        workflow.name = new_name

    if description:
        workflow.description = description

    if creators != NO_VALUE:
        workflow.creators = cast(List[Person], creators)

        if all(c.email != git_creator.email for c in workflow.creators):
            workflow.creators.append(git_creator)

    if keywords != NO_VALUE:
        workflow.keywords = cast(List[str], keywords)

    if isinstance(workflow, Plan):
        if custom_metadata:
            existing_metadata = [a for a in workflow.annotations if a.source != "renku"]

            existing_metadata.append(Annotation(id=Annotation.generate_id(), body=custom_metadata, source="renku"))

            workflow.annotations = existing_metadata

        workflow.set_parameters_from_strings(set_params)

        def _mod_params(workflow, changed_params, attr):
            for param_string in changed_params:
                name, new_value = param_string.split("=", maxsplit=1)
                new_value = new_value.strip(' "')

                found = False
                for collection in [workflow.inputs, workflow.outputs, workflow.parameters]:
                    for i, param in enumerate(collection):
                        if param.name == name:
                            new_param = param.derive(plan_id=workflow.id)
                            setattr(new_param, attr, new_value)
                            collection[i] = new_param
                            found = True
                            break
                    if found:
                        break
                else:
                    raise errors.ParameterNotFoundError(parameter=name, workflow=workflow.name)

        _mod_params(workflow, rename_params, "name")
        _mod_params(workflow, describe_params, "description")
    elif isinstance(workflow, CompositePlan) and len(map_params):
        workflow.set_mappings_from_strings(map_params)

    plan_gateway.add(workflow)
    return plan_view(workflow)


@inject.autoparams()
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def compose_workflow(
    name: str,
    description: Optional[str],
    mappings: Optional[List[str]],
    defaults: Optional[List[str]],
    links: Optional[List[str]],
    param_descriptions: Optional[List[str]],
    map_inputs: bool,
    map_outputs: bool,
    map_params: bool,
    link_all: bool,
    keywords: Optional[List[str]],
    steps: Optional[List[str]],
    sources: Optional[List[str]],
    sinks: Optional[List[str]],
    creators: Optional[List[Person]],
    activity_gateway: IActivityGateway,
    plan_gateway: IPlanGateway,
    project_gateway: IProjectGateway,
) -> CompositePlan:
    """Compose workflows into a CompositePlan.

    Args:
        name(str): Name of the new composed Plan.
        description(Optional[str]): Description for the Plan.
        mappings(Optional[List[str]]): Mappings between parameters of this and child Plans.
        defaults(Optional[List[str]]): Default values for parameters.
        links(Optional[List[str]]): Links between parameters of child Plans.
        param_descriptions(Optional[List[str]]): Descriptions of parameters.
        map_inputs(bool): Whether or not to automatically expose child inputs.
        map_outputs(bool): Whether or not to automatically expose child outputs.
        map_params(bool): Whether or not to automatically expose child parameters.
        link_all(bool): Whether or not to automatically link child steps' parameters.
        keywords(Optional[List[str]]): Keywords for the Plan.
        steps(Optional[List[str]]): Child steps to include.
        sources(Optional[List[str]]): Starting files when automatically detecting child Plans.
        sinks(Optional[List[str]]): Ending files when automatically detecting child Plans.
        creators(Optional[List[Person]]): Creator(s) of the composite plan.
        activity_gateway(IActivityGateway): Injected activity gateway.
        plan_gateway(IPlanGateway): Injected plan gateway.
        project_gateway(IProjectGateway): Injected project gateway.

    Returns:
        The newly created ``CompositePlan``.
    """
    from renku.command.view_model.composite_plan import CompositePlanViewModel
    from renku.core.workflow.activity import get_activities_until_paths, sort_activities

    if plan_gateway.get_by_name(name):
        raise errors.DuplicateWorkflowNameError(f"Duplicate workflow name: Workflow '{name}' already exists.")

    child_workflows = []
    plan_activities = []

    if steps:
        for workflow_name_or_id in steps:
            child_workflow = plan_gateway.get_by_id(workflow_name_or_id)

            if not child_workflow:
                child_workflow = plan_gateway.get_by_name(workflow_name_or_id)

            if not child_workflow or is_plan_removed(child_workflow):
                raise errors.ObjectNotFoundError(workflow_name_or_id)

            child_workflows.append(child_workflow)
    else:
        sources = sources or []
        sources = get_relative_paths(base=project_context.path, paths=sources)

        if not sinks:
            usages = activity_gateway.get_all_usage_paths()
            generations = activity_gateway.get_all_generation_paths()

            sinks = [g for g in generations if all(not are_paths_related(g, u) for u in usages)]

        sinks = get_relative_paths(base=project_context.path, paths=sinks)

        activities = list(get_activities_until_paths(sinks, sources, activity_gateway=activity_gateway))
        activities = sort_activities(activities)

        # we need to get the actual plans from the DB as plan_with_values returns a copy
        for i, activity in enumerate(activities, 1):
            child_workflow = activity.association.plan

            if is_plan_removed(child_workflow):
                raise errors.ParameterError(
                    f"The workflow '{child_workflow.name}' on activity '{activity.id}' is deleted."
                )
            child_workflows.append(child_workflow)
            plan_activities.append((i, activity.plan_with_values))

    if not creators:
        creators = [cast(Person, get_git_user(project_context.repository))]

    plan = CompositePlan(
        description=description,
        id=CompositePlan.generate_id(),
        keywords=keywords,
        name=name,
        plans=child_workflows,
        creators=creators,
        project_id=project_gateway.get_project().id,
    )

    if mappings:
        plan.set_mappings_from_strings(mappings)

    if defaults:
        plan.set_mapping_defaults(defaults)

    if plan_activities:
        # Since composite is created from activities, we need to add mappings to set defaults to the values of
        # the activities, to ensure values from the involved activities are preserved.
        # If the user supplies their own mappings, those overrule the automatically added ones.

        for i, child_plan in plan_activities:
            for param in itertools.chain(child_plan.inputs, child_plan.outputs, child_plan.parameters):
                try:
                    mapping_name = f"{i}-{param.name}"
                    plan.set_mappings_from_strings([f"{mapping_name}=@step{i}.{param.name}"])
                except errors.MappingExistsError:
                    continue

                plan.set_mapping_defaults([f"{mapping_name}={param.actual_value}"])

    if links:
        plan.set_links_from_strings(links)
        graph = ExecutionGraph([plan])
        cycles = graph.cycles
        if cycles:
            cycles = [map(lambda x: x.name, cycle) for cycle in cycles]
            raise errors.GraphCycleError(cycles)

    if param_descriptions:
        plan.set_mapping_descriptions(param_descriptions)

    if map_inputs:
        plan.map_all_inputs()

    if map_outputs:
        plan.map_all_outputs()

    if map_params:
        plan.map_all_parameters()

    if link_all:
        # NOTE: propagate values to for linking to use
        rv = CompositePlanValueResolver(plan)
        plan = rv.apply()

        graph = ExecutionGraph([plan], virtual_links=True)

        cycles = graph.cycles
        if cycles:
            cycles = [map(lambda x: x.name, cycle) for cycle in cycles]
            raise errors.GraphCycleError(cycles)

        for virtual_link in graph.virtual_links:
            plan.add_link(virtual_link[0], [virtual_link[1]])

    plan_gateway.add(plan)

    return CompositePlanViewModel.from_composite_plan(plan)


@inject.autoparams()
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def export_workflow(
    name_or_id,
    plan_gateway: IPlanGateway,
    format: str,
    output: Optional[Union[str, Path]],
    values: Optional[Dict[str, Any]],
    basedir: Optional[str],
    resolve_paths: Optional[bool],
    nest_workflows: Optional[bool],
):
    """Export a workflow to a given format.

    Args:
        name_or_id: name or id of the Plan to export
        plan_gateway(IPlanGateway): The injected Plan gateway.
        format(str): Format to export to.
        output(Optional[str]): Output path to store result at.
        values(Optional[Dict[str,Any]]): Parameter names and values to apply before export.
        basedir(Optional[str]): The base path prepended to all paths in the exported workflow,
            if None it defaults to the absolute path of the renku project.
        resolve_paths(Optional[bool]): Resolve all symlinks and make paths absolute, defaults to True.
        nest_workflows(Optional[bool]): Whether to try to nest all workflows into one specification and file or not,
            defaults to False.
    Returns:
        The exported workflow as string.
    """

    if resolve_paths is None:
        resolve_paths = True

    if basedir is None:
        basedir_path = project_context.path
    elif isinstance(basedir, str):
        basedir_path = Path(basedir)

    workflow = plan_gateway.get_by_name_or_id(name_or_id)

    if is_plan_removed(workflow):
        raise errors.ParameterError(f"The specified workflow '{name_or_id}' cannot be found.")

    output_path: Optional[Path] = None
    if output and isinstance(output, str):
        output_path = Path(output)
    elif output:
        output_path = cast(Path, output)

    if values:
        rv = ValueResolver.get(workflow, values)
        workflow = rv.apply()
        if rv.missing_parameters:
            communication.warn(
                f'Could not resolve the following parameters in "{workflow.name}" workflow: '
                f'{",".join(rv.missing_parameters)}'
            )

    from renku.core.plugin.workflow import workflow_converter

    converter = workflow_converter(format)
    return converter(
        workflow=workflow,
        basedir=basedir_path,
        output=output_path,
        output_format=format,
        resolve_paths=resolve_paths,
        nest_workflows=nest_workflows,
    )


def _lookup_paths_in_paths(lookup_paths: List[str], target_paths: List[str]):
    """Return all lookup_paths that are in or under target_paths."""

    dirs = []
    files = set()

    for p in lookup_paths:
        path = Path(get_relative_paths(base=project_context.path, paths=[p])[0])
        if path.is_dir():
            dirs.append(path)
        else:
            files.add(path)

    target_dirs = []
    target_files = set()

    for p in target_paths:
        path = Path(p)
        if path.is_dir():
            target_dirs.append(path)
        else:
            target_files.add(path)

    result = set()

    for target_file in target_files:
        if target_file in files or any(d in target_file.parents for d in dirs):
            result.add(str(target_file))

    for target_dir in target_dirs:
        if target_dir in dirs or any(target_dir in f.parents for f in files):
            result.add(str(target_dir))

    return result


@inject.autoparams()
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def visualize_graph(
    sources: List[str],
    targets: List[str],
    show_files: bool,
    activity_gateway: IActivityGateway,
    revision: Optional[str] = None,
):
    """Visualize an activity graph.

    Args:
        sources(List[str]): Input paths to start the visualized graph at.
        targets(List[str]): Output paths to end the visualized graph at.
        show_files(bool): Whether or not to show file nodes.
        activity_gateway(IActivityGateway): The injected activity gateway.
        revision(Optional[str], optional): Revision or revision range to show
            the graph for  (Default value = None)

    Returns:
        Graph visualization view model.
    """
    from renku.core.workflow.activity import create_activity_graph, get_activities_until_paths

    sources = sources or []
    sources = get_relative_paths(base=project_context.path, paths=[Path.cwd() / p for p in sources])

    if not targets:
        usages = activity_gateway.get_all_usage_paths()
        generations = activity_gateway.get_all_generation_paths()

        targets = [g for g in generations if all(not are_paths_related(g, u) for u in usages)]
    else:
        targets = get_relative_paths(base=project_context.path, paths=[Path.cwd() / p for p in targets])

    activities = get_activities_until_paths(
        paths=targets, sources=sources, revision=revision, activity_gateway=activity_gateway
    )
    graph = create_activity_graph(list(activities), with_inputs_outputs=show_files, with_hidden_dependencies=True)
    return ActivityGraphViewModel(graph)


@inject.autoparams()
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def workflow_inputs(activity_gateway: IActivityGateway, paths: Optional[List[str]] = None):
    """Get inputs used by workflows.

    Args:
        activity_gateway(IActivityGateway): The injected activity gateway.
        paths(List[str], optional): List of paths to consider as inputs (Default value = None).

    Returns:
        Set[str]: Set of input file paths.
    """
    usage_paths = activity_gateway.get_all_usage_paths()

    if not paths:
        return usage_paths

    return _lookup_paths_in_paths(lookup_paths=paths, target_paths=usage_paths)


@inject.autoparams()
@validate_arguments(config=dict(arbitrary_types_allowed=True))
def workflow_outputs(activity_gateway: IActivityGateway, paths: Optional[List[str]] = None):
    """Get inputs used by workflows.

    Args:
        activity_gateway(IActivityGateway): The injected activity gateway.
        paths(List[str], optional): List of paths to consider as outputs (Default value = None).

    Returns:
        Set[str]: Set of output file paths.
    """
    generation_paths = activity_gateway.get_all_generation_paths()

    if not paths:
        return generation_paths

    return _lookup_paths_in_paths(lookup_paths=paths, target_paths=generation_paths)


@inject.autoparams()
def get_initial_id(plan: Optional[AbstractPlan], plan_gateway: IPlanGateway) -> Optional[str]:
    """Return the id of the first plan in the derivative chain."""
    if plan is None:
        return None

    if not plan.derived_from:
        return plan.id

    parent = plan_gateway.get_by_id(plan.derived_from)
    if parent is None:
        raise errors.ParameterError(f"Cannot find parent plan with id '{plan.derived_from}'")

    return get_initial_id(parent)


@inject.autoparams()
def get_activities(plan: Optional[AbstractPlan], activity_gateway: IActivityGateway) -> Generator[Activity, None, None]:
    """Return all valid activities that use the plan or one of its parent/child derivatives."""
    if plan is None:
        return

    derivative_ids = [p.id for p in get_derivative_chain(plan=plan)]

    for activity in activity_gateway.get_all_activities():
        if activity.deleted:
            continue

        if activity.association.plan.id in derivative_ids:
            yield activity


def is_plan_removed(plan: AbstractPlan) -> bool:
    """Return true if the plan or any plan in its derivative chain is deleted."""
    for derived_plan in get_derivative_chain(plan):
        if derived_plan.deleted:
            return True

    return False


@inject.autoparams()
def get_composite_plans_by_child(plan: AbstractPlan, plan_gateway: IPlanGateway) -> List[CompositePlan]:
    """Return all composite plans that contain a child plan."""

    derivatives = {p.id for p in get_derivative_chain(plan=plan)}

    composites = (p for p in plan_gateway.get_newest_plans_by_names().values() if isinstance(p, CompositePlan))

    composites_containing_child = [c for c in composites if {p.id for p in c.plans}.intersection(derivatives)]

    return composites_containing_child


@inject.autoparams("activity_gateway", "plan_gateway")
def get_plans_with_metadata(activity_gateway: IActivityGateway, plan_gateway: IPlanGateway) -> List[AbstractPlan]:
    """Get all plans in the project with additional metadata.

    Adds information about last execution, number of executions and whether the plan was used to create files
    currently existing in the project.
    """

    all_activities = activity_gateway.get_all_activities()
    activity_map = _reverse_activity_plan_map(list(all_activities))
    latest_plan_chains: Set[Tuple[AbstractPlan]] = {
        cast(Tuple[AbstractPlan], tuple(get_derivative_chain(p)))
        for p in plan_gateway.get_newest_plans_by_names().values()
    }

    result: Dict[str, Union[Plan, CompositePlan]] = {}
    touches_file_cache: Dict[str, bool] = {}
    duration_cache: Dict[str, Optional[timedelta]] = {}

    # check which plans where involved in using/creating existing files
    for plan_chain in latest_plan_chains:
        latest_plan = cast(Union[Plan, CompositePlan], DynamicProxy(plan_chain[0]))
        latest_plan.touches_existing_files = False
        latest_plan.number_of_executions = 0
        latest_plan.created = latest_plan.date_created
        latest_plan.last_executed = cast(Optional[datetime], None)
        latest_plan.children = []
        duration = _get_plan_duration(latest_plan, duration_cache, activity_map)

        if duration is not None:
            latest_plan.duration = duration.seconds

        if isinstance(plan_chain[0], Plan):
            for activity in activity_map.get(latest_plan.id, []):
                if latest_plan.last_executed is None or latest_plan.last_executed < activity.ended_at_time:
                    latest_plan.last_executed = activity.ended_at_time

                latest_plan.number_of_executions += 1

                latest_plan.touches_existing_files = _check_workflow_touches_existing_files(
                    latest_plan, touches_file_cache, activity_map
                )
            latest_plan.type = "Plan"
        else:
            latest_plan.number_of_executions = None
            latest_plan.type = "CompositePlan"
            latest_plan.children = [get_latest_plan(p).id for p in latest_plan.plans]
            latest_plan.touches_existing_files = _check_workflow_touches_existing_files(
                latest_plan, touches_file_cache, activity_map
            )

        result[latest_plan.id] = latest_plan

    return list(result.values())  # type: ignore


def _reverse_activity_plan_map(activities: List[Activity], latest: bool = True) -> Dict[str, Set[Activity]]:
    """Create a map from plan id to relevant activities."""
    result: Dict[str, Set[Activity]] = {}

    for activity in activities:
        plan = activity.association.plan

        if latest:
            plan = get_latest_plan(plan)

        if plan.id not in result:
            result[plan.id] = {activity}
        else:
            result[plan.id].add(activity)

    return result


def _check_workflow_touches_existing_files(
    workflow: Union[Plan, CompositePlan],
    cache: Dict[str, bool],
    activity_map: Dict[str, Set[Activity]],
    latest: bool = True,
) -> bool:
    """Check if a workflow or one of its children touches existing files."""
    if latest:
        workflow = get_latest_plan(workflow)

    if workflow.id in cache:
        return cache[workflow.id]

    if isinstance(workflow, Plan):
        for activity in activity_map.get(workflow.id, []):
            for output in activity.generations:
                if (project_context.path / output.entity.path).exists():
                    return cache.setdefault(workflow.id, True)
    else:
        for child in workflow.plans:
            if _check_workflow_touches_existing_files(child, cache, activity_map):
                return cache.setdefault(workflow.id, True)

    return cache.setdefault(workflow.id, False)


def _get_plan_duration(
    workflow: Union[Plan, CompositePlan],
    cache: Dict[str, Optional[timedelta]],
    activity_map: Dict[str, Set[Activity]],
    latest: bool = True,
) -> Optional[timedelta]:
    if latest:
        workflow = get_latest_plan(workflow)

    if workflow.id in cache:
        return cache[workflow.id]

    if isinstance(workflow, Plan):
        times = []
        for activity in activity_map.get(workflow.id, []):
            times.append(activity.ended_at_time - activity.started_at_time)

        if not times:
            return cache.setdefault(workflow.id, None)

        return cache.setdefault(workflow.id, sum(times, timedelta(0)) / len(times))

    else:
        total = timedelta(0)
        found = False
        for child in workflow.plans:
            child_time = _get_plan_duration(child, cache, activity_map)

            if child_time is not None:
                total += child_time
                found = True

        if found:
            return cache.setdefault(workflow.id, total)

    return cache.setdefault(workflow.id, None)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku activity gateway interface."""

from abc import ABC
from pathlib import Path
from typing import List, Optional, Set, Tuple, Union

from renku.domain_model.provenance.activity import Activity, ActivityCollection


class IActivityGateway(ABC):
    """Interface for the ActivityGateway."""

    def get_by_id(self, id: str) -> Optional[Activity]:
        """Get an activity by id."""
        raise NotImplementedError

    def get_all_usage_paths(self) -> List[str]:
        """Return all usage paths."""
        raise NotImplementedError

    def get_all_generation_paths(self) -> List[str]:
        """Return all generation paths."""
        raise NotImplementedError

    def get_activities_by_usage(self, path: Union[Path, str], checksum: Optional[str] = None) -> List[Activity]:
        """Return the list of all activities that use a path."""
        raise NotImplementedError

    def get_activities_by_generation(self, path: Union[Path, str], checksum: Optional[str] = None) -> List[Activity]:
        """Return the list of all activities that generate a path."""
        raise NotImplementedError

    def get_downstream_activities(self, activity: Activity, max_depth=None) -> Set[Activity]:
        """Get downstream activities that depend on this activity."""
        raise NotImplementedError

    def get_upstream_activities(self, activity: Activity, max_depth=None) -> Set[Activity]:
        """Get upstream activities that this activity depends on."""
        raise NotImplementedError

    def get_downstream_activity_chains(self, activity: Activity) -> List[Tuple[Activity, ...]]:
        """Get a list of tuples of all downstream paths of this activity."""
        raise NotImplementedError

    def get_upstream_activity_chains(self, activity: Activity) -> List[Tuple[Activity, ...]]:
        """Get a list of tuples of all upstream paths of this activity."""
        raise NotImplementedError

    def get_all_activities(self, include_deleted: bool = False) -> List[Activity]:
        """Get all activities in the project."""
        raise NotImplementedError

    def add(self, activity: Activity) -> None:
        """Add an ``Activity`` to storage."""
        raise NotImplementedError

    def add_activity_collection(self, activity_collection: ActivityCollection):
        """Add an ``ActivityCollection`` to storage."""
        raise NotImplementedError

    def get_all_activity_collections(self) -> List[ActivityCollection]:
        """Get all activity collections in the project."""
        raise NotImplementedError

    def remove(self, activity: Activity, keep_reference: bool = True, force: bool = False):
        """Remove an activity from the storage.

        Args:
            activity(Activity): The activity to be removed.
            keep_reference(bool): Whether to keep the activity in the ``activities`` index or not.
            force(bool): Force-delete the activity even if it has downstream activities.
        """
        raise NotImplementedError



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Check your system and repository for potential problems.

.. cheatsheet::
   :group: Misc
   :command: $ renku doctor
   :description: Check your system and repository for potential problems.
   :target: rp

Commands and options
~~~~~~~~~~~~~~~~~~~~

.. rst-class:: cli-reference-commands

.. click:: renku.ui.cli.doctor:doctor
   :prog: renku doctor
   :nested: full
"""

import textwrap

import click

from renku.core import errors


@click.command()
@click.pass_context
@click.option("--fix", is_flag=True, help="Fix issues when possible.")
@click.option("-f", "--force", is_flag=True, help="Do possible fixes even though no problem is reported.")
def doctor(ctx, fix, force):
    """Check your system and repository for potential problems."""
    import renku.ui.cli.utils.color as color
    from renku.command.doctor import DOCTOR_INFO, doctor_check_command
    from renku.ui.cli.utils.callback import ClickCallback

    if force and not fix:
        raise errors.ParameterError("Cannot use '-f/--force' without '--fix'")

    click.secho("\n".join(textwrap.wrap(DOCTOR_INFO)) + "\n", bold=True)

    communicator = ClickCallback()
    command = doctor_check_command(with_fix=fix)
    if fix:
        command = command.with_communicator(communicator)
    is_ok, fixes_available, problems = command.build().execute(fix=fix, force=force).output

    if is_ok:
        click.secho("Everything seems to be ok.", fg=color.GREEN)
        ctx.exit(0)

    click.echo(problems)

    if fixes_available:
        click.echo("Run with '--fix' flag to try and fix these issues.")

    ctx.exit(1)
